import os, datetime, sys, types
from copy import deepcopy
import dateutil.parser

from numpy import cos, pi, cumsum, arange, ndarray, ones, zeros, array, newaxis, linspace, empty, resize, sin, allclose, zeros_like, linalg, dot, arctan2, float64, histogram2d, sum, sqrt, loadtxt, searchsorted, nan, logical_not, fliplr, flipud
import numpy
from numpy.ma import MaskedArray
import h5py

from reductus.dataflow.lib import hzf_readonly_stripped as hzf

from .FilterableMetaArray import FilterableMetaArray as MetaArray
from .he3analyzer import He3AnalyzerCollection

DEBUG=False

def hdf_to_dict(hdf_obj, convert_i1_tostr=True):
    out_dict = {}
    for key in hdf_obj:
        val = hdf_obj[key]
        if type(val) == h5py.highlevel.Dataset:
            if (val[()].dtype == 'int8') and (convert_i1_tostr == True):
                value = val[()].tostring()
            else:
                value = val[()]
            out_dict[key] = value
        else:
            out_dict[key] = hdf_to_dict(val)
    return out_dict

def LoadText(filename, friendly_name="", path=None, first_as_x=True):
    if path == None:
        path = os.getcwd()
    creation_story = "LoadText('{fn}')".format(fn=filename)
    data_in = loadtxt(os.path.join(path, filename))
    info = []
    first_y_col = 0
    if first_as_x:
        info.append({"name":"xaxis", "units":"unknown", "values": data_in[:,0]})
        first_y_col = 1
    else:
        info.append({"name":"rownumber", "units":"index", "values": range(data_in.shape[1])})

    info.append({"name":"Measurements", "cols":[]})
    for col in range(first_y_col, data_in.shape[1]):
        info[1]["cols"].append({"name": "column%d" % (col,)})

    info.append({"filename": filename, "start_datetime": None,
             "CreationStory":creation_story, "path":path})
    output_obj = MetaArray(data_in[:,first_y_col:], dtype='float', info=info[:])
    return output_obj


def LoadICPMany(filedescriptors):
    result = []
    for fd in filedescriptors:
        new_data = LoadICPData(fd.pop('filename'), **fd)
        if isinstance(new_data, list):
            result.extend(new_data)
        else:
            result.append(new_data)
    return result

DETECTOR_ACTIVE = (320, 340)

def LoadMAGIKPSD(filename, path="", friendly_name="", collapse_y=True, auto_PolState=False, PolState='', flip=True, transpose=True, **kw):
    """
    loads a data file into a MetaArray and returns that.
    Checks to see if data being loaded is 2D; if not, quits

    Need to rebin and regrid if the detector is moving...
    """
    lookup = {"DOWN_DOWN":"_down_down", "UP_DOWN":"_up_down", "DOWN_UP":"_down_up", "UP_UP":"_up_up", "entry": ""}
    if '.nxz' in filename:
        file_obj = hzf.File(filename)
    else:
        # nexus
        file_obj = h5py.File(os.path.join(path, filename))

    #if not (len(file_obj.detector.counts.shape) == 2):
        # not a 2D object!
    #    return
    for entryname, entry in file_obj.items():
        active_slice = slice(None, DETECTOR_ACTIVE[0], DETECTOR_ACTIVE[1])
        counts_value = entry['DAS_logs']['areaDetector']['counts'][:, 1:DETECTOR_ACTIVE[0]+1, :DETECTOR_ACTIVE[1]]
        dims = counts_value.shape
        print(dims)
        ndims = len(dims)
        if auto_PolState:
            PolState = lookup.get(entryname, "")
        # force PolState to a regularized version:
        if not PolState in lookup.values():
            PolState = ''
        #datalen = file_obj.detector.counts.shape[0]
        if ndims == 2:
            if DEBUG: print("2d")
            ypixels = dims[0]
            xpixels = dims[1]
        elif ndims >= 3:
            if DEBUG: print("3d")
            frames = dims[0]
            xpixels = dims[1]
            ypixels = dims[2]

        creation_story = "LoadMAGIKPSD('{fn}', path='{p}')".format(fn=filename, p=path, aPS=auto_PolState, PS=PolState)

        # doesn't really matter; changing so that each keyword (whether it took the default value
        # provided or not) will be defined
        #    if not PolState == '':
        #        creation_story += ", PolState='{0}'".format(PolState)
        # creation_story += ")"


        if ndims == 2: # one of the dimensions has been collapsed.
            info = []
            info.append({"name": "xpixel", "units": "pixels", "values": arange(xpixels) }) # reverse order
            info.append({"name": "theta", "units": "degrees", "values": entry['DAS_logs']['sampleAngle']['softPosition'][()] })
            info.extend([
                    {"name": "Measurements", "cols": [
                            {"name": "counts"},
                            {"name": "pixels"},
                            {"name": "monitor"},
                            {"name": "count_time"}]},
                    {"PolState": PolState, "filename": filename, "start_datetime": dateutil.parser.parse(file_obj.attrs.get('file_time')), "friendly_name": friendly_name,
                     "CreationStory":creation_story, "path":path, "det_angle":entry['DAS_logs']['detectorAngle']['softPosition'][()]}]
                )
            data_array = zeros((xpixels, ypixels, 4))
            mon =  entry['DAS_logs']['counter']['liveMonitor'][()]
            count_time = entry['DAS_logs']['counter']['liveTime'][()]
            if ndims == 2:
                mon.shape = (1,) + mon.shape # broadcast the monitor over the other dimension
                count_time.shape = (1,) + count_time.shape
            counts = counts_value
            if transpose == True: counts = counts.swapaxes(0,1)
            if flip == True: counts = flipud(counts)
            data_array[..., 0] = counts
            #data_array[..., 0] = file_obj.detector.counts
            data_array[..., 1] = 1
            data_array[..., 2] = mon
            data_array[..., 3] = count_time
            # data_array[:,:,4]... I wish!!!  Have to do by hand.
            data = MetaArray(data_array, dtype='float', info=info)
            data.friendly_name = friendly_name # goes away on dumps/loads... just for initial object.

        elif ndims == 3: # then it's an unsummed collection of detector shots.  Should be one sample and detector angle per frame
            if collapse_y == True:
                info = []
                info.append({"name": "xpixel", "units": "pixels", "values": arange(xpixels) }) # reverse order
                info.append({"name": "theta", "units": "degrees", "values": entry['DAS_logs']['sampleAngle']['softPosition'][()] })
                info.extend([
                        {"name": "Measurements", "cols": [
                                {"name": "counts"},
                                {"name": "pixels"},
                                {"name": "monitor"},
                                {"name": "count_time"}]},
                        {"PolState": PolState, "filename": filename, "start_datetime": dateutil.parser.parse(file_obj.attrs.get('file_time')), "friendly_name": friendly_name,
                         "CreationStory":creation_story, "path":path, "det_angle":entry['DAS_logs']['detectorAngle']['softPosition'][()]}]
                    )
                data_array = zeros((xpixels, frames, 4))
                mon =  entry['DAS_logs']['counter']['liveMonitor'][()]
                count_time = entry['DAS_logs']['counter']['liveTime'][()]
                if ndims == 3:
                    mon.shape = (1,) + mon.shape # broadcast the monitor over the other dimension
                    count_time.shape = (1,) + count_time.shape
                counts = numpy.sum(counts_value, axis=2)
                if transpose == True: counts = counts.swapaxes(0,1)
                if flip == True: counts = flipud(counts)
                data_array[..., 0] = counts
                #data_array[..., 0] = file_obj.detector.counts
                data_array[..., 1] = 1
                data_array[..., 2] = mon
                data_array[..., 3] = count_time
                # data_array[:,:,4]... I wish!!!  Have to do by hand.
                data = MetaArray(data_array, dtype='float', info=info)
                data.friendly_name = friendly_name # goes away on dumps/loads... just for initial object.
            else: # make separate frames
                infos = []
                data = []
                samp_angle =  entry['DAS_logs']['sampleAngle']['softPosition'][()]
                if samp_angle.shape[0] == 1:
                    samp_angle = numpy.ones((frames,)) * samp_angle
                det_angle = entry['DAS_logs']['detectorAngle']['softPosition'][()]
                if det_angle.shape[0] == 1:
                    det_angle = numpy.ones((frames,)) * det_angle
                for i in range(frames):
                    samp_angle =  entry['DAS_logs']['sampleAngle']['softPosition'][i]
                    det_angle = entry['DAS_logs']['detectorAngle']['softPosition'][i]
                    info = []
                    info.append({"name": "xpixel", "units": "pixels", "values": range(xpixels) })
                    info.append({"name": "ypixel", "units": "pixels", "values": range(ypixels) })
                    info.extend([
                        {"name": "Measurements", "cols": [
                                {"name": "counts"},
                                {"name": "pixels"},
                                {"name": "monitor"},
                                {"name": "count_time"}]},
                        {"PolState": PolState, "filename": filename, "start_datetime": entry['start_time'][()], "friendly_name": friendly_name,
                         "CreationStory":creation_story, "path":path, "samp_angle": samp_angle, "det_angle": det_angle}]
                    )
                    data_array = zeros((xpixels, ypixels, 4))
                    mon =  entry['DAS_logs']['counter']['liveMonitor'][i]
                    count_time = entry['DAS_logs']['counter']['liveTime'][i]
                    counts = counts_value[i]
                    if flip == True: counts = flipud(counts)
                    data_array[..., 0] = counts
                    data_array[..., 1] = 1
                    data_array[..., 2] = mon
                    data_array[..., 3] = count_time
                    # data_array[:,:,4]... I wish!!!  Have to do by hand.
                    subdata = MetaArray(data_array, dtype='float', info=info)
                    subdata.friendly_name = friendly_name + ("_%d" % i) # goes away on dumps/loads... just for initial object.
                    data.append(subdata)
    return data

def LoadICPData(filename, path="", friendly_name="", auto_PolState=False, PolState='', flip=True, transpose=True, **kw):
    """
    loads a data file into a MetaArray and returns that.
    Checks to see if data being loaded is 2D; if not, quits

    Need to rebin and regrid if the detector is moving...
    """
    lookup = {"a":"_down_down", "b":"_up_down", "c":"_down_up", "d":"_up_up", "g": ""}
    file_obj = load(os.path.join(path, filename), format='NCNR NG-1')
    dims = file_obj.detector.counts.shape
    ndims = len(dims)
    #if not (len(file_obj.detector.counts.shape) == 2):
        # not a 2D object!
    #    return
    if auto_PolState:
        key = friendly_name[-2:-1] # na1, ca1 etc. are --, nc1, cc1 are -+...
        PolState = lookup.get(key, "")
    # force PolState to a regularized version:
    if not PolState in lookup.values():
        PolState = ''
    #datalen = file_obj.detector.counts.shape[0]
    if ndims == 2:
        if DEBUG: print("2d")
        ypixels = file_obj.detector.counts.shape[0]
        xpixels = file_obj.detector.counts.shape[1]
    elif ndims >= 3:
        if DEBUG: print("3d")
        frames = file_obj.detector.counts.shape[0]
        ypixels = file_obj.detector.counts.shape[1]
        xpixels = file_obj.detector.counts.shape[2]

    creation_story = "LoadICPData('{fn}', path='{p}', auto_PolState={aPS}, PolState='{PS}')".format(fn=filename, p=path, aPS=auto_PolState, PS=PolState)

    # doesn't really matter; changing so that each keyword (whether it took the default value
    # provided or not) will be defined
    #    if not PolState == '':
    #        creation_story += ", PolState='{0}'".format(PolState)
    # creation_story += ")"


    if ndims == 2: # one of the dimensions has been collapsed.
        info = []
        info.append({"name": "xpixel", "units": "pixels", "values": arange(xpixels) }) # reverse order
        info.append({"name": "theta", "units": "degrees", "values": file_obj.sample.angle_x })
        info.extend([
                {"name": "Measurements", "cols": [
                        {"name": "counts"},
                        {"name": "pixels"},
                        {"name": "monitor"},
                        {"name": "count_time"}]},
                {"PolState": PolState, "filename": filename, "start_datetime": file_obj.date, "friendly_name": friendly_name,
                 "CreationStory":creation_story, "path":path, "det_angle":file_obj.detector.angle_x}]
            )
        data_array = zeros((xpixels, ypixels, 4))
        mon = file_obj.monitor.counts
        count_time = file_obj.monitor.count_time
        if ndims == 2:
            mon.shape = (1,) + mon.shape # broadcast the monitor over the other dimension
            count_time.shape = (1,) + count_time.shape
        counts = file_obj.detector.counts
        if transpose == True: counts = counts.swapaxes(0,1)
        if flip == True: counts = flipud(counts)
        data_array[..., 0] = counts
        #data_array[..., 0] = file_obj.detector.counts
        data_array[..., 1] = 1
        data_array[..., 2] = mon
        data_array[..., 3] = count_time
        # data_array[:,:,4]... I wish!!!  Have to do by hand.
        data = MetaArray(data_array, dtype='float', info=info)
        data.friendly_name = friendly_name # goes away on dumps/loads... just for initial object.

    elif ndims == 3: # then it's an unsummed collection of detector shots.  Should be one sample and detector angle per frame
        infos = []
        data = []
        for i in range(frames):
            samp_angle = file_obj.sample.angle_x[i]
            det_angle = file_obj.detector.angle_x[i]
            info = []
            info.append({"name": "xpixel", "units": "pixels", "values": range(xpixels) })
            info.append({"name": "ypixel", "units": "pixels", "values": range(ypixels) })
            info.extend([
                {"name": "Measurements", "cols": [
                        {"name": "counts"},
                        {"name": "pixels"},
                        {"name": "monitor"},
                        {"name": "count_time"}]},
                {"PolState": PolState, "filename": filename, "start_datetime": file_obj.date, "friendly_name": friendly_name,
                 "CreationStory":creation_story, "path":path, "samp_angle": samp_angle, "det_angle": det_angle}]
            )
            data_array = zeros((xpixels, ypixels, 4))
            mon = file_obj.monitor.counts[i]
            count_time = file_obj.monitor.count_time[i]
            counts = file_obj.detector.counts[i]
            if flip == True: counts = flipud(counts)
            data_array[..., 0] = counts
            data_array[..., 1] = 1
            data_array[..., 2] = mon
            data_array[..., 3] = count_time
            # data_array[:,:,4]... I wish!!!  Have to do by hand.
            subdata = MetaArray(data_array, dtype='float', info=info)
            subdata.friendly_name = friendly_name + ("_%d" % i) # goes away on dumps/loads... just for initial object.
            data.append(subdata)
    return data




"""
Annulus mask
============

Create an antialiased annulus mask using PIL image commands.
"""

import numpy

def annular_mask_antialiased_pillow(shape, center, inner_radius, outer_radius,
                             background_value=0.0, mask_value=1.0,
                             oversampling=8):
    # type: (Tuple[int, int], Tuple[float, float], float, float, float, float, int)  -> numpy.ndarray
    """
    Takes the following:

    * *shape* tuple: (x, y) - this is the size of the output image
    * *center* tuple: (x, y)
    * *inner_radius*: float
    * *outer_radius*: float
    * *background_value*: float (the image is initialized to this value)
    * *mask_value*: float (the annulus is drawn with this value)
    * *oversampling*: int (the mask is drawn on a canvas this many times bigger
      than the final size, then resampled down to give smoother edges)
    """
    from PIL import Image, ImageDraw
    # Create a 32-bit float image
    intermediate_shape = (shape[0]*int(oversampling), shape[1]*int(oversampling))
    im = Image.new('F', intermediate_shape, color=background_value)

    # Making a handle to the drawing tool
    draw = ImageDraw.Draw(im)

    # Have to scale everything in the problem by the oversampling
    outer_radius_r = outer_radius * oversampling
    inner_radius_r = inner_radius * oversampling
    center_r = (center[0] * oversampling, center[1] * oversampling)


    # Calculate bounding box for outer circle
    x_outer_min = center_r[0] - outer_radius_r
    x_outer_max = center_r[0] + outer_radius_r
    y_outer_min = center_r[1] - outer_radius_r
    y_outer_max = center_r[1] + outer_radius_r
    outer_bbox = [x_outer_min, y_outer_min, x_outer_max, y_outer_max]

    # Calculate bounding box for inner circle
    x_inner_min = center_r[0] - inner_radius_r
    x_inner_max = center_r[0] + inner_radius_r
    y_inner_min = center_r[1] - inner_radius_r
    y_inner_max = center_r[1] + inner_radius_r
    inner_bbox = [x_inner_min, y_inner_min, x_inner_max, y_inner_max]

    # Draw the circles:  outer one first
    draw.ellipse(outer_bbox, fill=mask_value)

    # Now overlay the inner circle
    draw.ellipse(inner_bbox, fill=background_value)

    # Now bring it back to size, with antialiasing
    #im.thumbnail(shape, Image.ANTIALIAS)
    # This produced artifacts - output.max() was > mask_value by 10% or more!

    # Using numpy reshape instead (rebinning) - see Scipy cookbook
    output = numpy.asarray(im)
    output = output.reshape(shape[0], oversampling, shape[1], oversampling).mean(1).mean(2)
    return output

def rectangular_mask_antialiased_pillow(shape, rectangle_xy, background_value=0.0, mask_value=1.0, oversampling=8):
    # Create a 32-bit float image
    from PIL import Image, ImageDraw
    intermediate_shape = (shape[0]*int(oversampling), shape[1]*int(oversampling))
    im = Image.new('F', intermediate_shape, color=background_value)

    # Making a handle to the drawing tool
    draw = ImageDraw.Draw(im)

    x0,y0,x1,y1 = rectangle_xy

    draw.rectangle([x0*oversampling, y0*oversampling, x1*oversampling, y1*oversampling], fill=mask_value)
    output = numpy.asarray(im).T
    output = output.reshape(shape[0], oversampling, shape[1], oversampling).mean(1).mean(2)
    return output

def annular_ellipse_mask_antialiased_cairo(shape, center, inner_radius, outer_radius,
                        background_value=0.0, mask_value=1.0, compression_factor=1.0,
                        oversampling=1):

    import numpy
    import cairo
    surface = cairo.ImageSurface(cairo.FORMAT_A8, shape[1], shape[0])
    ctx = cairo.Context(surface)
    ctx.set_antialias(cairo.ANTIALIAS_GRAY)
    ctx.scale(1.0 / compression_factor, 1.0)

    # Arc(cx, cy, radius, start_angle, stop_angle)
    thickness = outer_radius - inner_radius
    avg_radius = (outer_radius + inner_radius)/2.0
    ctx.arc(center[1] * compression_factor, center[0], avg_radius, 0, 2*numpy.pi)
    ctx.close_path()

    ctx.set_source_rgba(1,1,1,1)  # Solid color
    ctx.set_line_width(thickness)
    ctx.stroke()
    buf = surface.get_data()
    data = numpy.ndarray(shape=shape, dtype=numpy.uint8, buffer=buf)
    output = numpy.ones_like(data, dtype=float) * background_value
    output += data.astype(float) * (mask_value - background_value) / 255.0
    return output

def rectangular_mask_antialiased_cairo(shape, rectangle_xy, background_value=0.0, mask_value=1.0, oversampling=8):
    # Create a 32-bit float image
    import numpy
    import cairo
    surface = cairo.ImageSurface(cairo.FORMAT_A8, shape[1], shape[0])
    ctx = cairo.Context(surface)
    ctx.set_antialias(cairo.ANTIALIAS_GRAY)
    ctx.set_source_rgba(0,0,0,0)  # Solid color
    ctx.paint()

    x0,y0,x1,y1 = rectangle_xy
    w = x1 - x0
    h = y1 - y0

    #ctx.set_source(cairo.SolidPattern(0.5, 0.5, 0.5))
    ctx.set_source_rgba(1.0,1.0,1.0,1.0)  # Solid color
    ctx.rectangle(y0, x0, h, w)
    ctx.fill()

    buf = surface.get_data()
    data = numpy.ndarray(shape=shape, dtype=numpy.uint8, buffer=buf)
    output = numpy.ones_like(data, dtype=float) * background_value
    output += data.astype(float) * (mask_value - background_value) / 255.0
    return output

def pie_bounding_box(center, radius, start_angle, end_angle):
    # assumes Euclidean angles (CCW) - make sure to specify pieslice with negative angles (CW)
    points = [
        center,
        [center[0] + numpy.cos(start_angle)*radius, center[1] + numpy.sin(start_angle)*radius],
        [center[0] + numpy.cos(end_angle)*radius, center[1] + numpy.sin(end_angle)*radius]
    ]
    pangle = numpy.pi/2.0 # angle of the cardinal points
    for cardinal in numpy.arange(pangle*numpy.ceil(start_angle/pangle), end_angle, pangle):
        points.append([center[0] + numpy.cos(cardinal)*radius, center[1] + numpy.sin(cardinal)*radius])

    points = numpy.array(points, dtype='float')
    xmin = points[:,0].min()
    xmax = points[:,0].max()
    ymin = points[:,1].min()
    ymax = points[:,1].max()
    return [xmin, ymin, xmax, ymax]

def sector_cut_antialiased(shape, center, inner_radius, outer_radius, start_angle=0.0, end_angle=numpy.pi,
                mirror=True, background_value=0.0, mask_value=1.0,
                oversampling=8):
    # type: (Tuple[int, int], Tuple[float, float], float, float, float, float, int)  -> numpy.ndarray
    """
    Takes the following:

    * *shape* tuple: (x, y) - this is the size of the output image
    * *center* tuple: (x, y)
    * *inner_radius*: float
    * *outer_radius*: float
    * *start_angle*: float (radians) start of sector cut range
    * *end_angle*: float (radians) end of sector cut range (with defaults, covers full circle)
    * *mirror*: bool (take cut on both sides of origin)
    * *background_value*: float (the image is initialized to this value)
    * *mask_value*: float (the annulus is drawn with this value)
    * *oversampling*: int (the mask is drawn on a canvas this many times bigger
      than the final size, then resampled down to give smoother edges)
    """
    d_start = -numpy.degrees(start_angle)
    d_end = -numpy.degrees(end_angle)

    # Create a 32-bit float image
    intermediate_shape = (shape[0]*int(oversampling), shape[1]*int(oversampling))
    im = Image.new('F', intermediate_shape, color=background_value)

    # Making a handle to the drawing tool
    draw = ImageDraw.Draw(im)

    # Have to scale everything in the problem by the oversampling
    outer_radius_r = outer_radius * oversampling
    inner_radius_r = inner_radius * oversampling
    center_r = (center[0] * oversampling, center[1] * oversampling)

    # Calculate bounding box for outer circle
    x_outer_min = center_r[0] - outer_radius_r
    x_outer_max = center_r[0] + outer_radius_r
    y_outer_min = center_r[1] - outer_radius_r
    y_outer_max = center_r[1] + outer_radius_r
    outer_bbox = [x_outer_min, y_outer_min, x_outer_max, y_outer_max]

    # draw pie slices
    draw.pieslice(outer_bbox, d_end, d_start, fill=mask_value)
    if mirror:
        draw.pieslice(outer_bbox, d_end-180.0, d_start-180.0,  fill=mask_value)

    # Calculate bounding box for inner circle
    x_inner_min = center_r[0] - inner_radius_r
    x_inner_max = center_r[0] + inner_radius_r
    y_inner_min = center_r[1] - inner_radius_r
    y_inner_max = center_r[1] + inner_radius_r
    inner_bbox = [x_inner_min, y_inner_min, x_inner_max, y_inner_max]

    # Now overlay the inner circle
    draw.ellipse(inner_bbox, fill=background_value)

    # Now bring it back to size, with antialiasing
    #im.thumbnail(shape, Image.ANTIALIAS)
    # This produced artifacts - output.max() was > mask_value by 10% or more!

    # Using numpy reshape instead (rebinning) - see Scipy cookbook
    output = numpy.asarray(im)
    output = output.reshape(shape[0], oversampling, shape[1], oversampling).mean(1).mean(2)
    return output

def check_sector_cut_antialiased(shape, center, inner_radius, outer_radius, start_angle=0.0, end_angle=numpy.pi,
                mirror=True, background_value=0.0, mask_value=1.0,
                oversampling=8):
    # type: (Tuple[int, int], Tuple[float, float], float, float, float, float, int)  -> numpy.ndarray
    """
    Takes the following:

    * *shape* tuple: (x, y) - this is the size of the output image
    * *center* tuple: (x, y)
    * *inner_radius*: float
    * *outer_radius*: float
    * *start_angle*: float (radians) start of sector cut range
    * *end_angle*: float (radians) end of sector cut range (with defaults, covers full circle)
    * *background_value*: float (the image is initialized to this value)
    * *mask_value*: float (the annulus is drawn with this value)
    * *oversampling*: int (the mask is drawn on a canvas this many times bigger
      than the final size, then resampled down to give smoother edges)
    """
    d_start = -numpy.degrees(start_angle)
    d_end = -numpy.degrees(end_angle)

    # Create a 32-bit float image
    intermediate_shape = (shape[0]*int(oversampling), shape[1]*int(oversampling))
    im = Image.new('F', intermediate_shape, color=background_value)

    # Making a handle to the drawing tool
    draw = ImageDraw.Draw(im)

    # Have to scale everything in the problem by the oversampling
    outer_radius_r = outer_radius * oversampling
    inner_radius_r = inner_radius * oversampling
    center_r = (center[0] * oversampling, center[1] * oversampling)

    # Calculate bounding box for outer circle
    x_outer_min = center_r[0] - outer_radius_r
    x_outer_max = center_r[0] + outer_radius_r
    y_outer_min = center_r[1] - outer_radius_r
    y_outer_max = center_r[1] + outer_radius_r
    outer_bbox = [x_outer_min, y_outer_min, x_outer_max, y_outer_max]
    draw.pieslice(outer_bbox, d_end, d_start, fill=mask_value)
    if mirror:
        draw.pieslice(outer_bbox, d_end-180.0, d_start-180.0,  fill=mask_value)

    # Calculate bounding box for inner circle
    x_inner_min = center_r[0] - inner_radius_r
    x_inner_max = center_r[0] + inner_radius_r
    y_inner_min = center_r[1] - inner_radius_r
    y_inner_max = center_r[1] + inner_radius_r
    inner_bbox = [x_inner_min, y_inner_min, x_inner_max, y_inner_max]

    # Now overlay the inner circle
    draw.ellipse(inner_bbox, fill=background_value)
    im.show()

# prefer cairo version because it doesn't require oversampling, and is GPU accelerated!
try:
    import cairo
    annular_mask_antialiased = annular_ellipse_mask_antialiased_cairo
    rectangular_mask_antialiased = rectangular_mask_antialiased_cairo
except ImportError:
    annular_mask_antialiased = annular_mask_antialiased_pillow
    rectangular_mask_antialiased = rectangular_mask_antialiased_pillow
def check():
    check_sector_cut_antialiased(
        (128,128),
        (64, 60),
        10,
        20,
        start_angle=numpy.pi/4,
        end_angle=numpy.pi/2,
        mirror=False,
        background_value=0.0,
        mask_value=100.0,
        oversampling=8
    )

if __name__ == "__main__":
    check()



"""
Polarized data corrections

Polarized neutron measurements select a particular neutron spin state
for the incident and reflected neutrons.  Experiments which probe the
interaction between spin state and sample must first determine the
efficiency with which each spin state can be selected and apply that
effect to the data before presenting it to the user.[1]

The way that spin states are selected will vary between instruments.
Some instruments will have a polarizing supermirror+flipper
arrangement, with the mirror either transmitting or reflecting
a particular polarization state, and a magnetic field tuned to
flip the polarization state when a current is applied or otherwise
leave it in the selected state.  TOF sources will require a
time-varying field to adjust for the different transit times
for different wavelengths through the flipper field.  He3 polarizers
can select spin up or spin down for transmission and so do not
require a flipper.  In any case, the efficiency will vary with
Q, either due to increased divergence or wavelength dependence.


1.  Prepare the intensity data
==============================

To perform a polarization data reduction you must first have measured
beam data under different polarization conditions::

    pp spin up incident, spin up reflected
    pm spin up incident, spin down measured
    mp spin down incident, spin up measured
    mm spin down incident, spin down measured

These data are passed into the correction using a PolarizedData container::

    I = PolarizedData()
    I.xlabel, P.xunits = 'Slit 1', 'mm'   # On TOF, maybe 'wavelength','nm'
    I.ylabel, P.yunits = 'Intensity', 'counts'
    I.pp.set(x=Spp, y=Ipp, variance=Ipp)
    I.pm.set(x=Spm, y=Ipm, variance=Ipm)
    I.mp.set(x=Smp, y=Imp, variance=Imp)
    I.mm.set(x=Smm, y=Imm, variance=Imm)

The intensities given should be 1-dimensional.  For 1D and 2D detectors, the
integrated beam image should be used.  Time-of-flight data will be
organized by time channel while monochromatic data is organized by slit
opening.  Any ordering will do so long as the intensity is well behaved
between points.

We will assume the correction is uniform across the detector since it
is impractical to measure the efficiency as a function of position.
In fact the slightly different path lengths for the various positions
on the detector will lead to a decrease in efficiency as you move away
from the center of the beam, and so it may have a subtle influence on
off-specular polarized reflectometry.

The different cross sections must have corresponding ordinate axes.  On TOF
machines this will usually not be a problem since all time channels are
measured simultaneously.  On scanning instruments, there are a variety
of factors which may result in points that don't correspond well, such
as user error, or undersampling in the spin-flip cross sections.

Even if the data are already aligned across the four cross sections,
it is still desirable to smooth the polarized data prior to estimating
because the efficiency estimate is unstable.

Use alignment or smoothing corrections to make your data regular::

    I.apply(align)    # use linear interpolation to regrid the data
    I.apply(smooth)   # a weighted irregular Savitsky-Golay filter


2. Compute polarization efficiencies
====================================

The polarization efficiency correction is controlled by the
PolarizationEfficiency class::

    eff = PolarizationEfficiency(I)

Once the class is constructed and the resulting efficiency attributes
are available::

    eff.fp front polarizer efficiency
    eff.ff front flipper efficiency
    eff.rp rear polarizer efficiency
    eff.rf rear flipper efficiency
    eff.Ic overall beam intensity

The formalism from Majkrzak, et al. (see attached PDF) defines the
following, which are attributes to the efficiency object::

    eff.F = fp
    eff.R = rp
    eff.x = 1 - 2*ff
    eff.y = 1 - 2*rf
    eff.beta = Ic/2

There are a few attributes of the class that will change how the
efficiencies are calculated::

    eff.min_intensity = 1e-2
    eff.min_efficiency = 0.7
    eff.FRbalance = 0.5

FRbalance determines the relative front-back weighting of the
polarization efficiency.  From the data we can only estimate the
product FR of the front and rear polarization efficiencies, and
the user must decide how to distribute the remainder.
The FRbalance should vary between 0 (front polarizer is 100% efficient)
through 0.5 (distribute inefficiency equally) to 1 (rear polarizer
is 100% efficient).  The particular formula used is::

     F = (F*R)^FRbalance
     R = (F*R)/F

The min_intensity and min_efficiency numbers are used to bound
the range of the correction.  Normally you won't need to set
them.  These are class attributes, so you can set them globally
in PolarizationEfficiency instead of eff, but still override
them for particular instances by setting them in eff.


3. Apply the correction to the data
===================================

You must first prepare your data for the polarization correction by
placing it in a PolarizedData container, and possibly subtracting
the background measurements (the polarization correction is linear so it
doesn't matter mathematically if background subtraction happens before
or after the polarization correction, but the polarization correction
is slow enough that it should probably only be done once).

Once again the measurements need to be aligned, but in this case
smoothing should not be used.  The data will need to be normalized
to the same monitor as the intensity measurement before adding them
to the container.

For scanning instruments, the data will need to retain the slit
settings used for the measurement so that the correct intensity
measurement can be used for the correction.


TODO: create a complete data reduction doc including footprint, etc.
TODO: use array masks to identify interpolated data.
TODO: incorporate time for He3 polarizer
TODO: extend to 1D and 2D detectors
TODO: cross check results against Asterix algorithm
TODO: implement alignment and smoothing
TODO: consider applying the efficiency correction to the theory rather than
the data --- if the inversion is unstable, this may be more reliable.
TODO: with both spin up and spin down neutrons in sample simultaneously
in some proportion, do the cross sections need to be coherently to account
for the theory?

[1] C.F. Majkrzak (1996). Physica B 221, 342-356.
"""
from __future__ import print_function, division

import numpy as np

# reflred.uncertainty is fast but overestimates the error for the highly
# correlated polarization efficiency estimate.  The uncertainties package
# from pypi is slow, and does an excellent job of matching monte carlo
# estimates of polarization efficiency.

#from reductus.dataflow.lib.uncertainty import Uncertainty
#def U(x,dx): return Uncertainty(x,dx**2)
#def nominal_values(u): return u.x
#def std_devs(u): return u.dx
from uncertainties.unumpy import uarray as U, nominal_values, std_devs, ulinalg
from uncertainties import ufloat

from reductus.dataflow.lib.errutil import interp

from . import util

# A divergence difference of 5e-5 corresponds to 0.1 mm for 2 m slit separation
ACCEPTABLE_DIVERGENCE_DIFFERENCE = 5e-5

ALL_XS = '++', '+-', '-+', '--'
PM_XS = '++', '+-', '--'
MP_XS = '++', '-+', '--'
NSF_XS = '++', '--'

class PolarizationData:
    def __init__(self, intensity_data, Imin=0.0, Emin=0.0, FRbal=0.5, clip=True):
        self.beam = dict([(d.polarization, d) for d in intensity_data])
        self.FRbal = FRbal
        self.Emin = Emin
        self.Imin = Imin
        self.clip = clip

        self.warnings = []

    def apply(self, data, spinflip=True):
        correct(data, spinflip, self.beam,
                Imin=self.Imin, Emin=self.Emin, FRbal=self.FRbal,
                clip=self.clip)

    def plot(self):
        raise NotImplementedError("see polarization.py for plots we want to create")

    def get_metadata(self):
        eff = polarization_efficiency(self.beam, Imin=self.Imin, Emin=self.Emin,
                                      FRbal=self.FRbal, clip=self.clip)
        output = dict([(k, str(v)) for k, v in eff.items()])
        return output
        
    def get_plottable(self):
        return {"params": self.get_metadata(), "type": "params"}

def apply_polarization_correction(data, polarization, spinflip=True):
    polarization.apply(data, spinflip=spinflip)

def correct(data_in, spinflip, beam, Imin=0.0, Emin=0.0, FRbal=0.5, clip=False):
    dtheta = beam['++'].angular_resolution
    beta, fp, rp, x, y, mask \
        = _calc_efficiency(beam=beam, dtheta=dtheta,
                           Imin=Imin, Emin=Emin, FRbal=FRbal, clip=clip)

    data = dict([(d.polarization, d) for d in data_in])
    use_pm = spinflip and '+-' in data
    use_mp = spinflip and '-+' in data

    Hinv = _correction_matrix(beta, fp, rp, x, y, use_pm, use_mp)
    _apply_correction(data, dtheta, Hinv, use_pm, use_mp)

def _apply_correction(data, dtheta, Hinv, use_pm, use_mp):
    """Apply the efficiency correction in eff to the data."""

    # Identify active cross sections
    if use_pm and use_mp:
        parts = ALL_XS
    elif use_mp:
        parts = MP_XS
    elif use_pm:
        parts = PM_XS
    else:
        parts = NSF_XS

    # TODO: interpolation with mixed resolution
    # Interpolate data so that it aligns with ++.  If smoothing is
    # desired, apply the interpolated smoothing before calling polcor,
    # in which case the interpolation does nothing.
    assert parts[0] == '++'
    x = data['++'].Qz
    y = [U(data['++'].v, data['++'].dv)]
    for p in parts[1:]:
        px = data[p].Qz
        py = U(data[p].v, data[p].dv)
        y.append(interp(x, px, py, left=np.nan, right=np.nan))
    Y = np.vstack(y)
    Y = U(nominal_values(Y), std_devs(Y))

    # Look up correction matrix for each point using the ++ cross section
    correction_index = util.nearest(data['++'].angular_resolution, dtheta)

    # Apply the correction at each point
    X, dX = np.zeros(Y.shape), np.zeros(Y.shape)
    for point, polarization in enumerate(correction_index):
        x = Hinv[polarization] @ Y[:, point].T
        X[:, point], dX[:, point] = nominal_values(x).flat, std_devs(x).flat

    # Put the corrected intensities back into the datasets
    # interpolate back to the original Qz in that dataset:
    for k, xs in enumerate(parts):
        x = data[xs].Qz
        px = data['++'].Qz
        py = U(X[k, :], dX[k, :])
        y = interp(x, px, py, left=np.nan, right=np.nan)
        data[xs].v, data[xs].dv = nominal_values(y), std_devs(y)
        data[xs].vlabel = 'counts per incident count'
        data[xs].vunits = None


def _correction_matrix(beta, fp, rp, x, y, use_pm, use_mp):
    """
    Generate polarization correction matrices for each slit configuration *dT*.
    """
    Fp, Fm = 1+fp, 1-fp
    Rp, Rm = 1+rp, 1-rp
    Fp_x, Fm_x = 1+fp*x, 1-fp*x
    Rp_y, Rm_y = 1+rp*y, 1-rp*y

    # pylint: disable=bad-whitespace
    if use_pm and use_mp:
        H = np.array([
            [Fm_x*Rm_y, Fm_x*Rp_y, Fp_x*Rm_y, Fp_x*Rp_y],
            [Fm_x*Rm  , Fm_x*Rp  , Fp_x*Rm  , Fp_x*Rp  ],
            [Fm  *Rm_y, Fm  *Rp_y, Fp  *Rm_y, Fp  *Rp_y],
            [Fm  *Rm  , Fm  *Rp  , Fp  *Rm  , Fp  *Rp  ],
            ])
    elif use_pm:
        H = np.array([
            [Fm_x*Rm_y, (Fm_x*Rp_y + Fp_x*Rm_y), Fp_x*Rp_y],
            [Fm_x*Rm  , (Fm_x*Rp   + Fp_x*Rm  ), Fp_x*Rp  ],
            [Fm  *Rm  , (Fm  *Rp   + Fp  *Rm  ), Fp  *Rp  ],
        ])
    elif use_mp:
        H = np.array([
            [Fm_x*Rm_y, (Fm_x*Rp_y + Fp_x*Rm_y), Fp_x*Rp_y],
            [Fm  *Rm_y, (Fm  *Rp_y + Fp  *Rm_y), Fp  *Rp_y],
            [Fm  *Rm  , (Fm  *Rp   + Fp  *Rm  ), Fp  *Rp  ],
        ])
    else:
        H = np.array([
            [Fm_x*Rm_y, Fp_x*Rp_y  ],
            [Fm  *Rm  , Fp  *Rp    ],
            ])

    #print("H:", UM(H[:,:,0]), UM(H[:,:,0]).I)
    return [ulinalg.pinv(H[:, :, k]*Bk) for k, Bk in enumerate(beta)]

def plot_efficiency(beam, Imin=0.0, Emin=0.0, FRbal=0.5, clip=False):
    eff = polarization_efficiency(beam, Imin=Imin, Emin=Emin, FRbal=FRbal, clip=clip)
    from matplotlib import pyplot as plt
    ax1 = plt.subplot(211)
    for xs in ALL_XS:
        beam[xs].plot()
    _ploteff(eff, 'beta')
    plt.legend()
    plt.setp(ax1.get_xticklabels(), visible=False)
    plt.xlabel('')

    plt.subplot(212, sharex=ax1)
    for part in ('ff', 'rf', 'fp', 'rp'):
        _ploteff(eff, part)
    plt.grid(True)
    plt.legend()
    plt.ylabel("efficiency (%)")
    #plt.xlabel('angular resolution (degrees 1-sigma)')
    plt.xlabel('slit 1 opening (mm)')


EFF_LABELS = {'ff':'front flipper', 'rf':'rear flipper',
              'fp':'front polarizer', 'rp':'rear polarizer'}
EFF_SCALES = {'ff':100, 'rf':100, 'fp':100, 'rp':100}
EFF_COLORS = {'ff':'green', 'rf':'blue', 'fp':'cyan', 'rp':'purple'}
def _ploteff(eff, part):
    from matplotlib import pyplot as plt
    label = EFF_LABELS.get(part, part)
    scale = EFF_SCALES.get(part, 1.0)
    color = EFF_COLORS.get(part, 'black')
    x, mask = eff['slit1'], eff['mask']
    y, dy = scale*nominal_values(eff[part]), scale*std_devs(eff[part])
    plt.errorbar(x, y, dy, fmt='.', color=color, label=label, capsize=0, hold=True)
    #print "mask",mask
    #if np.any(mask):
        #plt.errorbar(x[mask], y[mask], dy[mask],
        #             fmt='.', color='red', label='_', capsize=0, hold=True)
        #plt.plot(x[mask], y[mask], '.', color='red', hold=True)
        #plt.vlines(x[mask], (y-dy)[mask], (y+dy)[mask], colors='red', hold=True)

def polarization_efficiency(beam, Imin=0.0, Emin=0.0, FRbal=0.5, clip=False):
    """
    Compute polarization and flipper efficiencies from the intensity data.

    *beam* is the measured beam intensity as a dict with entries for
    ++, +-, -+, and --.

    *FRbalance* is the  portion of efficiency assigned to front versus rear

    *clip* [True|False] force efficiencies below 100%

    If clip is true, constrain points between the minimum efficiency and one,
    and intensity above minimum intensity.

    The computed values are systematically related to the efficiencies:

    Ic: intensity is 2*beta
    fp: front polarizer efficiency is F
    rp: rear polarizer efficiency is R
    ff: front flipper efficiency is (1-x)/2
    rf: rear flipper efficiency is (1-y)/2
    mask: true if point was clipped.

    See PolarizationEfficiency.pdf for details on the calculation.
    """
    # Assume the '++' cross section exists and completely covers the
    # resolution range
    dtheta = beam['++'].angular_resolution
    s1 = beam['++'].slit1.x
    beta, fp, rp, x, y, mask = \
        _calc_efficiency(beam=beam, dtheta=dtheta, FRbal=FRbal,
                         Imin=Imin, Emin=Emin, clip=clip)
    ff, rf = (1-x)/2, (1-y)/2
    return dict(dtheta=dtheta, slit1=s1, beta=beta,
                ff=ff, rf=rf, fp=fp, rp=rp, mask=mask)

def _calc_efficiency(beam, dtheta, Imin, Emin, FRbal, clip):
    # Beam intensity.
    # NOTE: A:mm, B:pm, C:mp, D:mm
    #pp, pm, mp, mm = [_interp_intensity(dtheta, beam[xs]) for xs in ALL_XS]
    pp, pm, mp, mm = [_nearest_intensity(dtheta, beam[xs]) for xs in ALL_XS]

    Ic = ((mm*pp) - (pm*mp)) / ((mm+pp) - (pm+mp))
    reject = np.zeros_like(Ic, dtype='bool')  # Reject nothing initially
    if clip:
        reject |= _clip_data(Ic, Imin, np.inf)
    beta = Ic/2


    # F and R are the front and rear polarizer efficiencies.  Each
    # is limited below by min_efficiency and above by 1 (since they
    # are not neutron sources).  Keep a list of points that are
    # rejected because they are outside this range.
    FR = mm/Ic - 1
    if clip:
        reject |= _clip_data(FR, Emin**2, 1)
    fp = FR ** FRbal
    rp = FR / fp

    # f and r are the front and rear flipper efficiencies.  Each
    # is again limited below by min_efficiency and above by 1.
    # We don't compute f and r directly, but instead x, y, Fx and Fy:
    #    x = 1-2f => f=(1-x)/2
    #    y = 1-2r => r=(1-y)/2
    x = (pm/Ic - 1)/FR
    y = (mp/Ic - 1)/FR

    # Clip the f,r flipper efficiencies to [min. efficiency, 1] by clipping
    # x,y to [-1, 1-2*min. efficiency]
    if clip:
        reject |= _clip_data(x, -1, 1-2*Emin)
        reject |= _clip_data(y, -1, 1-2*Emin)

    return beta, fp, rp, x, y, reject

def _nearest_intensity(dT, data):
    index = util.nearest(dT, data.angular_resolution)
    if (abs(dT - data.angular_resolution[index]) > ACCEPTABLE_DIVERGENCE_DIFFERENCE).any():
        raise ValueError("polarization cross sections for direct beam are not aligned")
    return U(data.v, data.dv)[index]

def _interp_intensity(dT, data):
    return interp(dT, data.angular_resolution, U(data.v, data.dv))

# Different versions of clip depending on which uncertainty package is used.
def clip_no_error(field, low, high, nanval=0.):
    """
    Clip the values to the range, returning the indices of the values
    which were clipped.  Note that this modifies field in place. nan
    values are clipped to the nanval default.  *field* is a floating
    point array, with no uncertainty.
    """
    index = np.isnan(field)
    field[index] = nanval
    reject = index

    index = field < low
    field[index] = low
    reject |= index

    index = field > high
    field[index] = high
    reject |= index

    return reject


def clip_reflred_err1d(field, low, high, nanval=0.):
    """
    Clip the values to the range, returning the indices of the values
    which were clipped.  Note that this modifies field in place. nan
    values are clipped to the nanval default.

    *field* is dataflow.uncertainty array, whose values retain their
    variance even if they are forced within the bounds.  *low*, *high*
    and *nanval* are floats.
    """
    index = np.isnan(field.x)
    field[index] = U(nanval, field.variance[index])
    reject = index

    index = field.x < low
    field[index] = U(low, field.variance[index])
    reject |= index

    index = field.x > high
    field[index] = U(high, field.variance[index])
    reject |= index

    return reject


def clip_pypi_uncertainties(field, low, high, nanval=0.):
    """
    Clip the values to the range, returning the indices of the values
    which were clipped.  Note that this modifies field in place. nan
    values are clipped to the nanval default.

    *field* is an array from the uncertainties package, whose values retain
    their variance even if they are forced within the bounds.  Clipping is
    performed by subtracting from the uncertain value, which will help
    preserve correlations when performing further operations on the clipped
    values.

    *low*, *high* and *nanval* are floats.
    """
    # Move value to the limit without changing the correlated errors.
    # This is probably wrong, but it is less wrong than other straight forward
    # options, such setting x to the limit with zero uncertainty.  At least
    # the clipped points will be flagged.
    index = np.isnan(nominal_values(field))
    field[index] = [ufloat(nanval, 0.) for v in field[index]]
    reject = index

    index = field < low
    field[index] = [v+(low-v.n) for v in field[index]]
    reject |= index

    index = field > high
    field[index] = [v-(v.n-high) for v in field[index]]
    reject |= index

    return reject


if 'Uncertainty' in globals():
    _clip_data = clip_reflred_err1d
else:
    _clip_data = clip_pypi_uncertainties



# This program is public domain
"""
Background subtraction
"""
from __future__ import division

import numpy as np

from reductus.dataflow.lib.uncertainty import Uncertainty as U, interp

from .refldata import ReflData

def apply_specular_mask(data):
    theta_i = data.sample.angle_x
    theta_f = 0.5*data.detector.angle_x
    dtheta = 0.1*data.angular_resolution
    not_spec = (abs(theta_f - theta_i) >= dtheta)
    data.mask = not_spec | (True if data.mask is None else data.mask)
    return data


def apply_background_subtraction(data, backp, backm):
    assert all((bkg is None or data.normbase == bkg.normbase)
               for bkg in (backp, backm)), "can't mix time and monitor normalized data"
    I, varI = subtract_background(data, backp, backm)
    data.v, data.dv = I, np.sqrt(varI)
    return data

def apply_interpolation(data, base, align='auto'):
    # type: (ReflData, ReflData) -> None
    """
    Interpolates background measurements to align with the specular.  If a
    different interpolation is required, then do so before subtract_background
    so that interp does nothing herein.

    Note: modifies data, data.sample and data.detector
    """
    #print "%s - (%s+%s)/2"%(spec.name, (backp.name if backp else "none"), (backm.name if backm else "none"))
    if align == 'auto':
        align = data.Qz_basis
    x = get_sample_angle(base, align)
    xp = get_sample_angle(data, align)
    vp = U(data.v, data.dv**2)
    v = interp(x, xp, vp)
    data.v, data.dv = v.x, np.sqrt(v.variance)
    # TODO: do we care that this throws away the offset information?
    data.sample.angle_x = x
    data.detector.angle_x = 2*x
    data.sample.angle_x_target = x
    data.detector.angle_x_target = 2*x
    # TODO: protect against funky changes in resolution when interpolating
    data.angular_resolution = base.angular_resolution
    # Remember the chosen alignment strategy for when we do future operations,
    # such as binning or subtraction.
    data.Qz_basis = align
    return data

def set_background_alignment(data, offset):
    """
    Guess whether background is offset from sample angle or from detector angle.
    """
    if offset is None or offset == 'auto':
        # for auto, if Qz_target is set by the trajectory then use it
        if data.Qz_target is not None and not np.isnan(data.Qz_target).any():
            data.Qz_basis = 'target'
            return
        offset = guess_background_offset(data)

    #A, B, L = data.sample.angle_x, data.detector.angle_x, data.detector.wavelength
    #if offset == 'sample':
    #   data.Qz_target = 4*np.pi/L * np.sin(np.radians(A))
    #elif offset == 'detector':
    #   data.Qz_target = 4*np.pi/L * np.sin(np.radians(B)/2)
    #else:
    #   raise KeyError('unknown offset: must be sample|detector')
    data.Qz_basis = offset
    return data


# TODO: should only subtract items with the same angular resolution
def subtract_background(spec, backp, backm):
    # type: (ReflData, ReflData, ReflData) -> (np.ndarray, np.ndarray)
    """
    Subtract back+ and back- from spec.

    if *offset* is 'auto', guess alignment.

    Interpolates background measurements to align with the specular.  If a
    different interpolation is required, then do so before subtract_background
    so that interp does nothing herein.

    Returns I, varI
    """
    #print "%s - (%s+%s)/2"%(spec.name, (backp.name if backp else "none"), (backm.name if backm else "none"))
    spec_v = U(spec.v, spec.dv**2)
    backp_v = U(backp.v, backp.dv**2) if backp else None
    backm_v = U(backm.v, backm.dv**2) if backm else None

    #print "spec",spec_v,spec.v,spec.dv
    #if backp: print "back+",backp_v,backp.v,backp.dv
    #if backm: print "back-",backm_v,backm.v,backm.dv

    backp_v = interp(_ordinate(spec), _ordinate(backp), backp_v) if backp else None
    backm_v = interp(_ordinate(spec), _ordinate(backm), backm_v) if backm else None

    if backp and backm:
        spec_v -= (backp_v + backm_v)/2
    elif backp:
        spec_v -= backp_v
    elif backm:
        spec_v -= backm_v
    else:
        pass  # no background to subtract

    return spec_v.x, spec_v.variance

def _ordinate(data):
    """
    Determine interpolation ordinate for the dataset.  This is Qz for 1-D data
    but theta for n-D data (since Q varies across the detector).
    """
    if data.Qz.ndim == 1:
        return data.Qz
    return get_sample_angle(data, data.Qz_basis)

def get_sample_angle(data, align):
    # This is similar to code in Refldata.Qz ...
    if align == 'detector':
        return data.detector.angle_x/2
    if align == 'sample':
        return data.sample.angle_x
    if align == 'detector_target':
        return data.detector.angle_x_target/2
    if align == 'sample_target':
        return data.sample.angle_x_target
    # for specular data, actual and target use both sample and detector angle.
    if align == 'actual':
        return data.sample.angle_x
    if align == 'target':
        return data.sample.angle_x_target
    raise KeyError("qz basis must be one of [actual, target, detector, sample, detector_target, sample_target]")

def guess_background_offset(back):
    """
    Guess whether background is offset from sample or from detector.

    Note that we can only distinguish relative and constant offsets,
    not  whether it is offset from sample or detector, so we must rely on
    convention. If the offset is constant for each angle, then it is assumed
    to be a sample offset.  If the offset is proportional to the angle (and
    therefore offset/angle is constant), then it is assumed to be a detector
    offset. If neither condition is met, it is assumed to be a sample offset.
    """
    a3 = back.sample.angle_x
    a4 = back.detector.angle_x
    a3_from_a4 = a4/2.
    a4_from_a3 = a3*2.
    a4_from_a3[a4_from_a3 == 0.] = 1e-5  # large number, but not many
    #print "a3",a3
    #print "a3 - a3 from a4",a3 - a3_from_a4
    #print "a4 - a4 from a3",a4 - a4_from_a3
    #print "a4",a4
    #print "a4 from a3",a4_from_a3
    #print "(a4 - a4_from_a3)/a4_from_a3",(a4 - a4_from_a3)/a4_from_a3
    if _check_mostly_constant(a3 - a3_from_a4):
        # A3 absolute offset
        return 'sample'
    #elif _check_mostly_constant((a3 - a3_from_a4)/a3_from_a4):
    #    # A3 relative offset
    #    return 'sample'
    #elif _check_mostly_constant(a4 - a4_from_a3):
    #    # A4 absolute offset
    #    return 'detector'
    elif _check_mostly_constant((a4 - a4_from_a3)/a4_from_a3):
        # A4 relative offset
        return 'detector'
    else:
        return 'sample'


def _check_mostly_constant(v):
    # normalize
    # find median; don't want mean since it is not robust
    med = np.median(v)
    delta = abs(med)*0.05
    # exclude points too far away from central value
    #print med,delta,v
    outliers = np.sum((v < med-delta) | (v > med+delta))
    #print "outliers",outliers
    # if too many points excluded, then reject the "mostly constant" assumption
    return outliers <= len(v)//10


def test_alignment_guess():
    back = ReflData()
    #a3 = np.arange(0.005,3,0.005)
    #a3 = np.arange(0.5,3,0.5)
    a3 = np.arange(-0.5, 6, 0.5)
    a3[1] = 0.
    a4 = 2*a3
    a3err = np.random.uniform(-0.0001, 0.0001, size=a3.size)
    a4err = np.random.uniform(-0.0001, 0.0001, size=a3.size)

    # detector offset
    back.sample.angle_x = a3 + a3err
    back.detector.angle_x = a4 + 0.3 + a4err
    assert guess_background_offset(back) == 'sample', "detector absolute +"
    back.detector.angle_x = a4 - 0.3 + a4err
    assert guess_background_offset(back) == 'sample', "detector absolute -"
    back.detector.angle_x = a4*(1+0.4) + a4err
    assert guess_background_offset(back) == 'detector', "detector relative +"
    back.detector.angle_x = a4*(1-0.4) + a4err
    assert guess_background_offset(back) == 'detector', "detector relative -"

    # sample offset
    back.detector.angle_x = a4 + a4err
    back.sample.angle_x = a3 + 0.3 + a3err
    assert guess_background_offset(back) == 'sample', "sample absolute +"
    back.sample.angle_x = a3 - 0.3 + a3err
    assert guess_background_offset(back) == 'sample', "sample absolute -"
    back.sample.angle_x = a3*(1+0.4) + a3err
    assert guess_background_offset(back) == 'detector', "sample relative +"
    back.sample.angle_x = a3*(1-0.4) + a3err
    assert guess_background_offset(back) == 'detector', "sample relative -"



# This program is public domain
"""
Reflectometry data representation.

Need to support collections of data from TOF, monochromatic
and white beam instruments.

Conceptually each data point is a tuple::

    incident angles (sample tilt and rotation)
    reflected angles (polar angle of detector pixel)
    slit distances and openings
    detector pixel distance and size
    incident/reflected polarization
    wavelength distribution
    measurement start and duration
    monitor and detector counts
    sample environment

Reflectometers are either vertical or horizontal geometry.
For vertical geometry (sample surface parallel to gravity),
x refers to horizontal slit opening and horizontal detector
pixels.  For horizontal geometry (sample surface perpendicular
to gravity) x refers to vertical slit opening and vertical
detector pixels. Other than gravitational corrections to
resolution and detector pixels, the analysis for the two
instrument types should be identical.

Monochromatic reflectometry files have a single wavelength per
angle and a series of angles.  Time-of-flight and polychromatic
reflectometers have multiple wavelengths per angle but usually
one angle per file.  In either case a file is a set of
detector frames each with its own wavelength and angles.

Different polarization states will be treated as belonging
to different measurements.  These will need to be aligned
before polarization correction can be performed.  Multiple
measurements may occur on the same detector.  In this
case each measurement should have a separate 'region of
interest' to isolate it from the others, presenting a virtual
detector to the reduction and analysis program.

Some information about the measurements may be missing
from the files, or recorded incorrectly.  Changes and
additions to the metadata must be recorded in any reduced
data file format, along with a list of transformations
that went into the reduction.

See notes in properties.py regarding dated values.
"""
__all__ = ['ReflData']

## Proposal for introspection with units
#def get_as(object,**kw):
#    if len(kw) != 1: raise Error
#    for k,v in kw.items():
#        return convert(getattr(object,k),object.__units__[k],v)
#def units(object, k):
#    return object.__units__[k]
#
#get_as(detector, distance='m')
#units(detector, 'distance')
#
#class Detector(object):
#    __units__ = dict(distance='mm',size='mm',saturation='counts/s')
#    ...
#
## Variation: use a units class to represent the units rather than a string
## This means that we save a couple of string lookups when doing conversion
#get_as(detector, distance=metre)
#def get_as(object,**kw):
#    if len(kw) != 1: raise Error
#    for k,v in kw.items():
#        return getattr(object,k)*object.__units__[k]/v
#class Detector(object):
#    __units__ = dict(distance=milli*metre,size=milli*metre,saturation=1/second)

## Something similar can be used for uncertainty, preferably stored as variance

import sys
import datetime
import warnings
import json
from io import BytesIO

import numpy as np
from numpy import inf, arctan2, sqrt, sin, cos, pi, radians

from reductus.dataflow.lib.exporters import exports_text, exports_json, exports_HDF5, NumpyEncoder
from reductus.dataflow.lib.strings import _s, _b
from .resolution import calc_Qx, calc_Qz, dTdL2dQ

IS_PY3 = sys.version_info[0] >= 3

# for sample background angle offset
QZ_FROM_SAMPLE = 'sample angle'
QZ_FROM_DETECTOR = 'detector angle'

class Group(object):
    _fields = ()
    _props = ()
    def __setattr__(self, key, value):
        # Check for class attr when setting; this is because hasattr on
        # a property will return False if getattr on that property raises
        # an exception.  This means if you really want to sneak an
        # attribute into the group from your data loader, you will have
        # to populate it from the
        if not key.startswith('_') and not hasattr(self.__class__, key):
            raise AttributeError("Cannot add attribute %s to class %s"
                                 % (key, self.__class__.__name__))
        object.__setattr__(self, key, value)
    def __init__(self, **kw):
        _set(self, kw)
    def __str__(self):
        return _str(self)
    def _toDict(self):
        return _toDict(self)

def set_fields(cls):
    groups = set(name for name, type in getattr(cls, '_groups', ()))
    properties = []
    fields = []
    for k, v in sorted((k, v) for k, v in cls.__dict__.items()):
        if k.startswith('_') or k in groups:
            pass
        elif isinstance(v, property):
            properties.append(k)
        elif not callable(v):
            fields.append(k)
    cls._fields = tuple(fields)
    cls._props = tuple(properties)
    return cls

# TODO: attribute documentation and units should be integrated with the
# TODO: definition of the attributes.  Value attributes should support
# TODO: unit conversion
@set_fields
class Slit(Group):
    """
    Define a slit for the instrument.  This is needed for correct resolution
    calculations and for ab initio footprint calculations.

    distance (inf millimetre)
        Distance from sample.  Positive numbers are after the sample,
        negative numbers are before the sample along the beam path.
    offset (4 x 0 millimetre)
        Offset of the slit blades relative to the distance from the sample.
        For vertical geometry, this is left, right, up, down.  For horizontal
        geometry this is up, down, left, right.  Offset + distance gives the
        distances of the individual blades from the sample, with negative
        numbers occurring before the sample position and positive numbers
        after.
    shape (shape='rectangular')
        Whether we have slit blades ('rectangular') or a circular
        aperature ('circular').
    x (n x inf millimetre)
        Slit opening in the primary direction.  For vertical geometry
        this is the horizontal opening, for horizontal geometry this is
        the vertical opening.  This may be a constant (fixed slits) or
        of length n for the number of measurements.
    y (n x inf millimetre)
        Slit opening in the secondary direction.  This may be a constant
        (fixed slits) or of length n for the number of measurements.
    """
    properties = ['distance','offset','x','y','shape']
    columns = {"x": {"units": "mm"}}
    distance = inf
    offset = (0., 0., 0., 0.)
    x = inf  # type: np.ndarray
    x_target = None  # type: np.ndarray
    y = inf  # type: np.ndarray
    y_target = None  # type: np.ndarray
    shape = "rectangular" # rectangular or circular


@set_fields
class Environment(Group):
    """
    Define sample environment data for the measurements such as
        temperature (kelvin)
        pressure (pascal)
        relative_humidity (%)
        electric_field (V/m)
        magnetic_field (tesla)
        stress_field (pascal)

    The data may be a constant, a series of values equal to
    the number of scan points, or a series of values and times.
    The average, max and min over all scan points, and the
    value, max and min for a particular scan point may be
    available.

    Some measurements are directional, and will have a polar
    and azimuthal angle associated with them.  This may be
    constant for the entire scan, or stored separately with
    each magnitude measurement.
    """

    #: Name of environment variable
    name = ""
    #: Units to report on graphs
    units = ""
    #: Statistics on all measurements
    average = None  # type: np.ndarray
    minimum = None  # type: np.ndarray
    maximum = None  # type: np.ndarray
    #: Magnitude of the measurement
    value = None  # type: np.ndarray
    #: Start time for log (seconds)
    start = None  # type: np.ndarray
    #: Measurement time relative to start (seconds)
    time = None  # type: np.ndarray


@set_fields
class Sample(Group):
    """
    Define the sample geometry.  Size and shape areneeded for correct
    resolution calculations and for ab initio footprint calculations.
    Angles are needed for correct calculation of Q.  Rotation and
    environment are for display to the user.

    description ("")
        Sample description, if available from the file.
    width (inf millimetre)
        Width of the sample in the primary direction.  For fixed slits
        the footprint of the beam on the sample decreases with angle
        in this direction.
    length (inf millimetre)
        Width of the sample in the secondary direction.  The footprint
        is independent of angle in this direction.
    thickness (inf millimetre)
        Thickness of the sample.
    substrate_sld (10^-6 Angstrom^-2)
        To plot Fresnel reflectivity we need to know the substrate
        scattering length density.  The default is to assume silicon.
    shape ('rectangular')
        Shape is 'circular' or 'rectangular'
    angle_x (n x 0 degree)
        Angle between neutron beam and sample surface in the primary
        direction.  This may be constant or an array of length n for
        the number of measurements.
    angle_x_target (n x 0 degree)
        Desired angle_x, used by join to select the points that are
        nominally the same in the joined data set.
    angle_y (n x 0 degree)
        Angle between the neutron beam and sample surface in the
        secondary direction.  This may be constant or an array of
        length n for the number of measurements.  This is known as
        tilt on some instruments.
    rotation (n x 0 degree)
        For off-specular reflectivity the orientation of the patterned
        array on the surface of the sample affects the computed theory.
        This value is not needed for data reduction, but it should be
        reported to the user during reduction and carried through to
        the reduced file for correct analysis.
    environment ({})
        Sample environment data.  See Environment class for a list of
        common environment data.
    """
    name = ''
    description = ''
    columns = {"angle_x": {"units": "degrees"}}
    width = inf  # mm
    length = inf  # mm
    thickness = inf # mm
    shape = 'rectangular' # rectangular or circular or irregular
    angle_x = 0. # degree
    angle_x_target = None  # degree
    angle_y = 0. # degree
    rotation = 0. # degree
    substrate_sld = 2.07 # inv A  (silicon substrate for neutrons)
    incident_sld = 0. # inv A (air)
    broadening = 0.
    environment = None  # type: Dict[str, Environment]
    temp_setpoint = None
    temp_avg = None
    magnet_setpoint = None
    magnet_avg = None

    def __init__(self, **kw):
        self.environment = {}
        Group.__init__(self, **kw)


@set_fields
class Beamstop(Group):
    """
    Define the geometry of the beamstop.  This is used by the
    detector class to compute the shadow of the beamstop on the
    detector.  The beamstop is assumed to be centered on the
    direct beam regardless of the position of the detector.

    distance (0 millimetre)
        Distance from sample to beamstop.  Note: this will need to
        be subtracted from the distance from detector to beamstop.
    shape ('rectangular')
        Shape is 'circular' or 'rectangular'
    width (0 millimetre)
        Width of the beamstop in the primary direction.  For circular
        beamstops, this is the diameter.
    length (0 millimetre)
        Width of the beamstop in the secondary direction.  For circular
        beamstops, this is the diameter.
    offset (2 x millimetre)
        Offset of the beamstop from the center of the beam.
    ispresent (False)
        True if beamstop is present in the experiment.
    """
    distance = 0. # mm
    width = 0. # mm
    length = 0. # mm
    shape = 'rectangular' # rectangular or circular
    offset = (0., 0.) # mm
    ispresent = False


@set_fields
class Monochromator(Group):
    """
    Monochromator properties.
    wavelength (k nanometre)
        Wavelength for each channel
    wavelength_resolution (k %)
        Wavelength resolution of the beam for each channel using 1-sigma
        gaussian approximation dL, expressed as 100*dL/L.  The actual
        wavelength distribution is considerably more complicated, being
        approximately square for multi-sheet monochromators and highly
        skewed on TOF machines.
    """
    columns = {
        "wavelength": {"units": "Angstroms", "variance": "wavelength_resolution"},
    }
    wavelength = None # angstrom
    wavelength_resolution = None # angstrom


@set_fields
class Detector(Group):
    """
    Define the detector properties.  Note that this defines a virtual
    detector.  The real detector may have e.g., multiple beam paths
    incident upon it, and be split into two virtual detectors when
    the file is loaded.

    Direction x refers to the primary direction and y refers to
    the secondary direction.  For vertical geometry, the primary
    direction is in the horizontal plane and the secondary direction
    is in the vertical plane.  For horizontal geometry these are
    reversed.  This allows the reduction software to be simpler,
    but may complicate file loading from formats which store values
    in absolute geometry.

    Geometry
    ========
    dims (2 x pixels)
        Dimensions of the detector, [nx,ny].  For pencil detectors this
        should be [1,1].  For position sensitive detectors, this should be
        [nx,1].  For area detectors, this should be [nx,ny].
    distance (millimetre)
        Distance from the sample to the detector.
    size (2 x millimetre)
        Detector size, [x,y].  Default is 1 mm x 1 mm.
    solid_angle (2 x radian)
        Detector solid angle [x,y], calculated from distance and size.
    center (2 x millimetre)
        Location of the center pixel [x,y] relative to the detector arm.
    width_x (nx x millimetre)
    width_y (ny x millimetre)
        Pixel width in x and y.
    offset_x (nx x millimetre)
    offset_y (ny x millimetre)
        Pixel offset in x and y.
    angle_x (n x degree)
    angle_y (n x degree)
        Angle of the detector arm relative to the main beam in x and y.
        This may be constant or an array of length n for the number of
        measurements in the scan.
    angle_x_offset (nx x degree)
    angle_y_offset (ny x degree)
        Pixel angle relative to detector angle in x and y.
    rotation (degree)
        Angle of rotation of the detector relative to the beam.  This
        will affect how vertical integration in the region of interest
        is calculated.  Ideally the detector would not be rotated, though
        misalignment can sometimes occur.

    Efficiency
    ==========
    efficiency (nx x ny %)
        Efficiency of the individual pixels; this is an array of the same
        shape as the detector, giving the relative efficiency of each pixel,
        or 1 if the efficiency is unknown.
        TODO: do we need variance?
    saturation (k [rate (counts/second), efficiency (%), uncertainty])
        Given a measurement of a given number of counts versus expected
        number of counts on the detector (e.g., as estimated by scanning
        a narrow slit across the detector to measure the beam profile,
        then measuring increasingly large portions of the beam profile),
        this can be converted to an efficiency correction per count rate
        which can be applied to all data read with this detector.  The
        value for deadtime should be a tuple of vectors: count rate,
        efficiency and uncertainty.  Below the lowest count rate the detector
        is considered to be 100% efficient (any baseline inefficiency will
        be normalized when comparing the measured reflection to the
        measured beam).  Beyond the highest count rate, the detector
        is considered saturated.  The normalize counts vector (v,dv) will
        be scaled by 1/(saturation +/- uncertainty).

        Note: There may be separate per pixel and per detector
        saturation levels.
    mask (nx x ny)
        Ignore data when (mask&0xFFFF != 0)
        https://manual.nexusformat.org/classes/base_classes/NXdetector.html

    Measurement
    ===========
    wavelength (k nanometre)
        Wavelength for each channel
    wavelength_resolution (k %)
        Wavelength resolution of the beam for each channel using 1-sigma
        gaussian approximation dL, expressed as 100*dL/L.  The actual
        wavelength distribution is considerably more complicated, being
        approximately square for multi-sheet monochromators and highly
        skewed on TOF machines.
    time_of_flight (k+1 millisecond)
        Time boundaries for time-of-flight measurement
    counts (nx x ny x k counts OR n x nx x ny counts OR n x nx x ny x k counts)
        nx x ny detector pixels
        n number of measurements
        k time-of-flight/wavelength channels
    counts_variance (like counts)
    """
    dims = (1, 1) # i,j
    distance = None # mm
    size = (1., 1.)  # mm
    center = (0., 0.) # mm
    width_x = 1. # mm
    width_y = 1. # mm
    offset_x = 0. # mm
    offset_y = 0. # mm
    angle_x = 0.  # degree
    angle_y = 0.  # degree
    angle_x_target = 0.  # degree
    angle_x_offset = 0. # degree
    angle_y_offset = 0. # degree
    rotation = 0. # degree
    efficiency = 1. # proportion
    saturation = inf # counts/sec
    wavelength = None # angstrom
    wavelength_resolution = None # angstrom
    time_of_flight = None  # ms
    counts = None
    counts_variance = None
    mask = None
    deadtime = None
    deadtime_error = None
    columns = {
        "counts": {"units": "counts", "variance": "counts_variance"},
        "angle_x": {"units": "degrees"},
        "wavelength": {"units": "Angstroms", "variance": "wavelength_resolution"},
    }

    @property
    def solid_angle(self):
        """Detector solid angle [x,y] (radians)"""
        #return 2*arctan2(np.asarray(self.size)/2., self.distance)
        return (2*arctan2(np.asarray(self.size)/2., self.distance)
                if self.distance is not None
                else np.array([0., 0.]))


@set_fields
class ROI(Group):
    """
    Detector region of interest.

    Defines a rectangular region of interest on the detector which
    is used for defining frames.  This can be used for example to
    split a single detector with both polarization states (via
    transmission and reflection off a supermirror) into two virtual
    detectors.

    xlo, xhi (pixels)
    ylo, yhi (pixels)
    """
    xlo = None
    xhi = None
    ylo = None
    yhi = None

@set_fields
class Monitor(Group):
    """
    Define the monitor properties.

    The monitor is essential to the normalization of reflectometry data.
    Reflectometry is the number of neutrons detected divided by the
    number of neutrons incident on the sample.  To compute this ratio,
    the incident and detected neutrons must be normalized to the neutron
    rate, either counts per monitor count, counts per second or counts
    per unit of source power (e.g., coulombs of protons incident on the
    detector, or megawatt hours of reactor power).

    counts (n x k counts)
        Number of counts measured.  For scanning instruments there is
        a separate count for each of the n measurements.  For TOF
        instruments there is a separate count for each of k time
        channels.  Counts may be absent, in which case normalization
        must be by time or by monitor.  In some circumstances the
        user may generate a counts vector, for example by estimating
        the count rate by other means, in order to combine data
        measured by time with data measured by monitor when the
        monitor values are otherwise unreliable.
    counts_variance (n x k counts)
        Variance is set to the number of counts but scaled during
        monitor saturation and deadtime corrections.
    roi_counts (n x k counts)
        Count against a region of interest (ROI) on the detector.
        **TODO**: ROI is **not** scaled during detector deadtime corrections,
        and there is no correction for detector efficiency.
    roi_variance (n x k counts)
        Variance is to be the number of counts.
    count_time (n seconds)
        Duration of the measurement.  For scanning instruments, there is
        a separate duration for each measurement.  For TOF, this is a
        single value equal to the duration of the entire measurement.
    time_step (seconds)
        The count_time timer has a reporting unit, e.g. second, or
        millisecond, or in the case of NCNR ICP files, hundredths of
        a minute.  The measurement uncertainty for the count time
        is assumed to be uniform over the time_step, centered on
        the reported time, with a gaussian approximation of uncertainty
        being sqrt(time_step/12).
    source_power (n source_power_units)
        The average source power for each measurement. For situations when
        the monitor cannot be trusted (which can happen from time to time on
        some instruments), we can use the number of protons incident on
        the target (proton charge) or the energy of the source (reactor
        power integrated over the duration of each measurement) as a proxy
        for the monitor.
    source_power_units ('coulombs/s' | 'megawatts')
        Units for source power.
    source_power_variance (n source_power_units)
        Variance in the measured source power
    base ('time' | 'monitor' | 'roi' | 'power')
        The measurement rate basis which should be used to normalize
        the data.  This is initialized by the file loader, but may
        be overridden during reduction.
    start_time (n seconds)
        For scanning instruments the start of each measurement relative
        to start of the scan.  Note that this is not simply sum of the
        count times because there may be motor movement between
        measurements.  The start time is required to align the measurement
        values with environment parameters, and for calculation of He3
        polarization.  For TOF, this should be zero.
    distance (metre)
        Distance from the sample.  This is not used by reduction but
        may be of interest to the user.
    sampled_fraction ([0,1])
        Portion of the neutrons that are sampled by the monitor.  If the
        monitor is after the second slit, the monitor value can be used to
        estimate the the counts on the detector, scaled by the sampled
        fraction.   Otherwise a full slit scan is required to normalize
        the reflectivity.  This is the inverse of the detector to monitor
        ratio used to normalize data on some instruments.
    time_of_flight (k+1 millisecond)
        Time boundaries for the time-of-flight measurement

    The corrected monitor counts field will start as None, but may be
    set by a dead time correction, which scales the monitor according
    to the monitor rate.  If for some reason the monitor is flaky, then
    the corrected monitor counts could be set by multiplying time by
    the monitor rate.
    """
    distance = None
    sampled_fraction = None
    counts = None
    counts_variance = None
    roi_counts = None
    roi_variance = None
    start_time = None
    count_time = None
    time_step = 1 # Default to nearest second
    time_of_flight = None
    base = 'monitor'
    source_power = None # No source power recorded
    source_power_units = "MW"
    source_power_variance = 0
    saturation = None
    columns = {
        "counts": {"units": "counts", "variance": "counts_variance"},
        "roi_counts": {"units": "counts", "variance": "roi_variance"},
        "count_time": {"units": "seconds"},
        "source_power": {"units": source_power_units, "variance": "source_power_variance"}
    }
    deadtime = None
    deadtime_error = None

class Intent(object):
    """
    Intent is one of the following:

        intensity: Normalization scan for computing absolute reflection
        specular: Specular intensity measurement
        background+: Background measurement, sample rotated
        background-: Background measurement, detector offset
        rock qx: Rocking curve with fixed Qz
        rock sample: Rocking curve with fixed detector angle
        rock detector: Rocking curve with fixed sample angle
        unknown: Some other kind of measurement

        detector efficiency: Flood fill

    Not supported:
        alignment: Sample alignment measurement
        area: Measurement of a region of Qx-Qz plane
        slice: Slice through Qx-Qz
    """
    slit = 'intensity'
    spec = 'specular'
    back = 'background'
    backp = 'background+'
    backm = 'background-'
    rockQ = 'rock qx'
    rock3 = 'rock sample'
    rock4 = 'rock detector'
    none = 'unknown'
    deff = 'detector efficiency'
    time = 'time'
    other = 'other'
    scan = 'scan'

    intents = (slit, spec, back, backp, backm, rockQ, rock3, rock4,
               deff, time, other, scan, none)

    @staticmethod
    def isback(intent):
        return intent.startswith('background')

    @staticmethod
    def isspec(intent):
        return intent == Intent.spec

    @staticmethod
    def isrock(intent):
        return intent.startswith('rock')

    @staticmethod
    def isslit(intent):
        return intent == Intent.slit

    @staticmethod
    def isnone(intent):
        return intent == Intent.none

    @staticmethod
    def isscan(intent):
        return intent == Intent.scan

def infer_intent(data):
    """
    Infer intent from data.

    Returns one of the Intent strings.
    """
    # TODO: doesn't handle alignment scans
    theta_i = data.sample.angle_x
    theta_f = 0.5*data.detector.angle_x
    dtheta = 0.1*data.angular_resolution
    n = len(theta_i)

    scan_i = (max(theta_i) - min(theta_i) > dtheta).any()
    scan_f = (max(theta_f) - min(theta_f) > dtheta).any()
    if (abs(theta_i) < dtheta).all() and (abs(theta_f) < dtheta).all():
        # incident and reflected angles are both 0
        intent = Intent.slit
    elif (scan_i and scan_f) or (not scan_i and not scan_f):
        # both theta_i and theta_f are moving, or neither is moving
        if (abs(theta_f - theta_i) < dtheta).all():       # all specular
            intent = Intent.spec
        elif (data.Qz.max() - data.Qz.min() < data.dQ.max()
              and data.Qx.max() - data.Qx.min()) > data.dQ.max():
            intent = Intent.rockQ
        elif np.sum(theta_f - theta_i > dtheta) > 0.9*n:  # 90% above
            intent = Intent.backp
        elif np.sum(theta_i - theta_f > dtheta) > 0.9*n:  # 90% below
            intent = Intent.backm
        else:
            intent = Intent.none
    elif scan_i:
        # only theta_i is moving
        intent = Intent.rock3
    elif scan_f:
        # only theta_f is moving
        intent = Intent.rock4
    else:
        # never gets here
        intent = Intent.scan

    return intent

def exports_ORSO_text(name="column"):
    def inner_function(f):
        f.exporter = ORSO_text
        f.export_name = name
        return f
    return inner_function

def ORSO_text(datasets, export_method=None, template_data=None, concatenate=False):
    from io import StringIO
    from orsopy import fileio
    from reductus.dataflow.lib.exporters import _build_filename
    exports = [getattr(d, export_method)() for d in datasets]
    # exports should contain items of class OrsoDataset
    outputs = []
    if concatenate and exports:
        parts = [export['value'] for export in exports]
        for part in parts:
            part.info.reduction.software.name = "reductus"
            part.info.reduction.software.template_data = template_data['template_data']
        with StringIO() as output_buffer:
            fileio.save_orso(parts, output_buffer, data_separator="\n\n")
            output_buffer.seek(0)
            export_string = output_buffer.read()
        filename = _build_filename(exports[0], ext=".ort")
        outputs.append({"filename": filename, "value": export_string})
    else:
        for i, export in enumerate(exports):
            part = export['value']
            part.info.reduction.software.name = "reductus"
            part.info.reduction.software.template_data = template_data['template_data']
            with StringIO() as output_buffer:
                fileio.save_orso([part], output_buffer, data_separator="\n\n")
                output_buffer.seek(0)
                export_string = output_buffer.read()
            filename = _build_filename(export, ext=".ort", index=i)
            outputs.append({"filename": filename, "value": export_string})
    return outputs


@set_fields
class ReflData(Group):
    """
    Reflectometry data structure, giving a predictable name space for the
    reduction steps regardless of input file format.
    """

    _groups = (
        ("slit1", Slit), ("slit2", Slit), ("slit3", Slit), ("slit4", Slit),
        ("monochromator", Monochromator),
        ("detector", Detector),
        ("sample", Sample),
        ("monitor", Monitor),
        ("roi", ROI),
    )

    #: Sample geometry
    sample = None     # type: Sample
    #: Presample slits
    slit1 = None      # type: Slit
    #: Presample slits
    slit2 = None      # type: Slit
    #: Post sample slits
    slit3 = None      # type: Slit
    #: Post sample slits
    slit4 = None      # type: Slit
    #: Monochromator wavelength
    monochromator = None  # type: Monochromator
    #: Detector geometry, efficiency and counts
    detector = None   # type: Detector
    #: Counts and/or durations
    monitor = None    # type: Monitor
    #: Region of interest on the detector.
    roi = None        # type: ROI

    #: Name of a particular instrument
    instrument = "unknown"
    #: Whether the scattering plane is horizontal or vertical.  The x-y
    #: values of the slits, detector, etc. should be relative to the
    #: scattering plane, not the lab frame.
    geometry = "vertical"
    #: Type of radiation (neutron or xray) used to probe the sample.
    probe = "unknown"
    #: Location of the datafile
    path = "unknown"
    #: Download location for the file, if available
    uri = "unknown"
    #: For scanning instruments, the number of measurements.
    points = 1
    #: For time of flight, the number of time channels.  For white
    #: beam instruments, the number of analysers.
    channels = 1
    #: Name of the dataset.  This may be a combination of filename and
    #: entry number.
    name = ""
    #: Numeric ID of the dataset.  Using "fileNum" from trajectoryData
    filenumber = 0
    #: Entry identifier if more than one entry per file
    entry = ""
    #: Description of the entry.
    description = ""
    #: Starting date and time of the measurement.
    date = datetime.datetime(1970, 1, 1)
    #: Duration of the measurement.
    duration = 0
    #: Nominal attenuation as recorded in the data file, or 1.0 if not recorded
    attenuation = 1.0
    #: '' unpolarized
    #: '+' spin up
    #: '-' spin down
    #: '++','--'  non-spin-flip
    #: '-+','+-'  spin flip
    polarization = ""
    #: The base for normalization (e.g., 'monitor' or 'time')
    normbase = None
    #: List of warnings generated when the file was loaded
    warnings = None
    #: Value label for y-axis on 1-D or colorbar on 2-D plots.
    #: Label will change when the value is normalized.
    vlabel = 'Intensity'
    #: Value units
    vunits = 'counts'
    #: Value scale ('linear', 'log' or '' for auto)
    #: Units will change when the value is normalized.
    vscale = '' # type: str
    #: X axis label
    xlabel = 'unknown intent'
    #: X axis units
    xunits = ''
    #: Value scale ('linear', 'log' or '' for auto)
    xscale = ''
    #: points excluded from reduction
    mask = None
    #: Computed 1-sigma angular resolution in degrees
    angular_resolution = None  # type: np.ndarray
    #: Hint for axis to use for aligning data with intensity scans:
    align_intensity = 'angular_resolution'
    #: For background scans, the choice of Qz for the
    #: points according to theta (sample angle), 2theta (detector angle)
    #: or qz (Qz value of background computed from sample and detector angle)
    #: How to calculate Qz from instrument angles.
    #:
    #: **actual**
    #:     calculates Qx and Qz as (x,z)-components of
    #:     $(\vec k_{\text{out}} - \vec k_\text{in})$ in sample coordinates,
    #: **detector**
    #:     ignores the sample angle and calculates Qz
    #:     as $(4\pi/\lambda \sin(\theta_\text{detector}/2))$,
    #: **sample**
    #:     ignores the detector angle and calculates Qz
    #:     as $(4\pi/\lambda \sin(\theta_\text{sample}))$
    #: **target**
    #:     uses the user-supplied Qz_target values
    Qz_basis = 'actual'
    #: The target Qz value given in the data file; or NaN if not available
    Qz_target = None  # type: np.ndarray
    scan_value = None  # type: List[np.ndarray]
    scan_label = None  # type: List[str]
    scan_units = None  # type: List[str]

    _intent = Intent.none
    _v = None
    _dv = None

    ## Data representation for generic plotter as (x,y,z,v) -> (qz,qx,qy,Iq)
    ## TODO: subclass Data so we get pixel edges calculations
    #def _getx(self): return self.Qz
    #def _gety(self): return self.Qx
    #def _getz(self): return self.Qy
    #x,xlabel,xunits = property(_getx),"Qx","inv A"
    #y,ylabel,yunits = property(_gety),"Qy","inv A"
    #z,zlabel,zunits = property(_getz),"Qz","inv A"
    @property
    def intent(self):
        """Purpose of the measurement."""
        return self._intent

    @intent.setter
    def intent(self, v):
        # Note: not setting x value with the label since the returned x should
        # correspond to the underlying value even if it has been updated since
        # the measurement intent was set.
        self._intent = v
        if Intent.isspec(v) or Intent.isback(v):
            self.xlabel, self.xunits = "Qz", "1/Ang"
        elif Intent.isrock(v):
            self.xlabel, self.xunits = "Qx", "1/Ang"
        elif Intent.isslit(v):
            #self.xlabel, self.xunits = "angular resolution", "degrees 1-sigma"
            self.xlabel, self.xunits = "slit 1 opening", "mm"
        elif Intent.isscan(v):
            self.xlabel, self.xunits = self.scan_label[0], self.scan_units[0]
        else:
            self.xlabel, self.xunits = "point", ""

    @property
    def x(self):
        # Return different x depending on intent
        intent = self.intent
        if Intent.isback(intent) or Intent.isspec(intent):
            return self.Qz
        elif Intent.isrock(intent):
            return self.Qx
        elif Intent.isslit(intent):
            return self.slit1.x
            #return self.angular_resolution
        elif Intent.isscan(intent):
            return self.scan_value[0]
        else:
            return np.arange(1, len(self.v)+1)

    @property
    def dx(self):
        return self.dQ

    @property
    def v(self):
        return self.detector.counts if self._v is None else self._v

    @v.setter
    def v(self, v):
        self._v = v

    @property
    def dv(self):
        return sqrt(self.detector.counts_variance) if self._dv is None else self._dv

    @dv.setter
    def dv(self, dv):
        self._dv = dv

    @property
    def Ti(self):
        return self.sample.angle_x

    @property
    def Td(self):
        return self.detector.angle_x

    @property
    def Tf(self):
        Ti, Td = self.Ti, self.Td
        return Td - Ti if Ti is not None and Td is not None else None

    @property
    def Ti_target(self):
        return self.sample.angle_x_target

    @property
    def Td_target(self):
        return self.detector.angle_x_target

    @property
    def Tf_target(self):
        Ti, Td = self.Ti_target, self.Td_target
        return Td - Ti if Ti is not None and Td is not None else None

    @property
    def Li(self):
        return self.monochromator.wavelength

    @property
    def Ld(self):
        return self.detector.wavelength

    @property
    def dL(self):
        return self.detector.wavelength_resolution

    @property
    def Qz(self):
        # Note: specular reflectivity assumes elastic scattering
        Li = Ld = self.Ld
        #print("Qz_basis", self.Qz_basis, self.Ti.shape, self.Td.shape, self.Ti_target.shape, self.Td_target.shape, Li.shape)
        if self.Qz_basis == 'actual':
            return calc_Qz(self.Ti, self.Td, Li, Ld)
        if self.Qz_basis == 'target':
            if self.Qz_target is not None:
                return self.Qz_target
            return calc_Qz(self.Ti_target, self.Td_target, Li, Ld)
        # For background, can use detector angle or sample angle.
        # Note: sample angle won't work with CANDOR
        if self.Qz_basis == 'detector':
            return calc_Qz(self.Td/2, self.Td, Li, Ld)
        if self.Qz_basis == 'sample':
            return calc_Qz(self.Ti, 2*self.Ti, Li, Ld)
        if self.Qz_basis == 'detector_target':
            return calc_Qz(self.Td_target/2, self.Td_target, Li, Ld)
        if self.Qz_basis == 'sample_target':
            return calc_Qz(self.Ti_target, 2*self.Ti_target, Li, Ld)
        raise KeyError("Qz basis must be one of [actual, detector, sample, target]")

    @property
    def Qx(self):
        # Note: specular reflectivity assumes elastic scattering
        Li = Ld = self.Ld
        if Intent.isrock(self.intent) or self.Qz_basis == 'actual':
            return calc_Qx(self.Ti, self.Td, Li, Ld)
        if self.Qz_basis == 'target':
            return np.zeros_like(self.Td)
            #return calc_Qx(self.Ti_target, self.Td_target, Li, Ld)
        if self.Qz_basis == 'detector':
            return np.zeros_like(self.Td)
        if self.Qz_basis == 'sample':
            return np.zeros_like(self.Ti)
        raise KeyError("Qz basis must be one of [actual, detector, sample, target]")

    @property
    def dQ(self):
        if self.angular_resolution is None:
            return None
            #raise ValueError("Need to estimate divergence before requesting dQ")
        # TODO: move sample broadening to to the dQ calculation
        T, dT = self.Ti, self.angular_resolution
        L, dL = self.Ld, self.detector.wavelength_resolution
        #print(T.shape, dT.shape, L.shape, dL.shape)
        return dTdL2dQ(T, dT, L, dL)

    @property
    def columns(self):
        from copy import deepcopy
        from collections import OrderedDict
        data_columns = OrderedDict([
            ('x', {'label': self.xlabel, 'units': self.xunits, 'errorbars': 'dx'}),
            ('v', {'label': self.vlabel, 'units': self.vunits, 'errorbars': 'dv'}),
            ('Qz', {'label': 'Qz', 'units': "1/Ang"}),
            ('Qz_target', {'label': 'Target Qz', 'units': '1/Ang'}),
            ('Qx', {'label': 'Qx', 'units': "1/Ang"}),
            ('angular_resolution', {'label': 'Angular Resolution (1-sigma)', 'units': 'degrees'})
        ])
        # TODO: duplicate code in columns, apply_mask and refldata._group
        for subclsnm in ['sample', 'detector', 'monitor', 'slit1', 'slit2', 'slit3', 'slit4', 'monochromator']:
            subcls = getattr(self, subclsnm, None)
            if subcls is None:
                continue
            sub_cols = deepcopy(getattr(subcls, 'columns', {}))
            for col in sub_cols.keys():
                # units are defined for the subcolumns, but nothing else... do that here:
                sub_col = sub_cols[col]
                v = getattr(subcls, col, None)
                if v is not None and hasattr(v, 'size') and v.size > 0:
                    label = "%s/%s" % (subclsnm, col)
                    sub_col['label'] = label
                    data_columns[label] = sub_col
        if self.scan_value is not None:
            for si, sv in enumerate(self.scan_value):
                new_col = {}
                new_label = self.scan_label[si]
                new_col['label'] = new_label
                new_col['is_scan'] = True
                new_col['units'] = self.scan_units[si]
                data_columns[new_label] = new_col

        return data_columns

    def apply_mask(self, mask_indices):
        """in-place masking of all data that is maskable"""
        def check_array(v):
            return isinstance(v, np.ndarray)

        def make_mask(v, mask_indices):
            mask = np.ones_like(v, dtype="bool")
            mask[mask_indices] = False
            return mask

        for prop in ['_v', '_dv', 'angular_resolution', 'Qz_target']:
            v = getattr(self, prop, None)
            if check_array(v):
                # TODO: use numpy.delete(v, indices) instead
                masked_v = v[make_mask(v, mask_indices)]
                setattr(self, prop, masked_v)
                self.points = len(masked_v)

        self.scan_value = [v[make_mask(v, mask_indices)] if check_array(v) else v for v in self.scan_value]

        for subclsnm in ['sample', 'detector', 'monitor', 'slit1', 'slit2', 'slit3', 'slit4', 'monochromator']:
            subcls = getattr(self, subclsnm, None)
            if subcls is None:
                continue
            sub_cols = getattr(subcls, 'columns', {})
            for col in sub_cols.keys():
                v = getattr(subcls, col, None)
                if check_array(v):
                    setattr(subcls, col, v[make_mask(v, mask_indices)])
                # handle col_target
                target_name = col + "_target"
                v = getattr(subcls, target_name, None)
                if check_array(v):
                    setattr(subcls, target_name, v[make_mask(v, mask_indices)])
                # handle variance
                dv_name = sub_cols[col].get('variance', None)
                if dv_name is not None:
                    dv = getattr(subcls, dv_name, None)
                    if check_array(dv):
                        setattr(subcls, dv_name, dv[make_mask(dv, mask_indices)])
                rv_name = sub_cols[col].get('resolution', None)
                if rv_name is not None:
                    rv = getattr(subcls, rv_name, None)
                    if check_array(rv):
                        setattr(subcls, rv_name, rv[make_mask(rv, mask_indices)])

    def __init__(self, **kw):
        for attr, cls in self._groups:
            setattr(self, attr, cls())
        self.warnings = []
        Group.__init__(self, **kw)

    def __str__(self):
        base = [_str(self, indent=2)]
        others = ["".join(("  ", s, "\n", str(getattr(self, s))))
                  for s, _ in self._groups]
        return "\n".join(base+others)

    def todict(self, maxsize=np.inf):
        state = _toDict(self, maxsize=maxsize)
        groups = {s: _toDict(getattr(self, s), maxsize=maxsize)
                  for s, _ in self._groups}
        state.update(groups)
        return state

    def fromdict(self, state):
        props = {k: v for k, v in state.items() if k in self._fields}
        props = _fromDict(props)
        for k, v in props.items():
            setattr(self, k, v)
        for attr, cls in self._groups:
            props = _fromDict(state[attr])
            setattr(self, attr, cls(**props))

    def warn(self, msg):
        """Record a warning that should be displayed to the user"""
        warnings.warn(msg)
        self.warnings.append(msg)

    def get_metadata(self):
        """
        Return metadata used by webreduce.

        The following are used in webreduce/instruments/ncnr.refl.js::

            {
                x: [..., xmin, ..., xmax, ...]}
                sample: {name: str, description: str}
                intent: str
                polarization: str
                filenumber: int
                filename: str
                entryname: str
                mtime: int
                source: str name of data server uri
            }

        *x* min and max are used for drawing the range indicators.

        *sample.description* is displayed when hovering over link.

        *source* and *filename* are needed for creating the hdf reader link.

        *sample.name > intent > filenumber > polarization* forms the default
        tree ordering.

        Users can define their own tree organization from the other fields
        in the dataset, so we should probably include trajectoryData entries.
        Sample environment conditions could also be useful for some
        experiments.  In practice, though, the defaults are going to be
        good enough, and users won't be changing them.  Not sure what
        happens when a vector field is used as a sort criterion.
        """
        # TODO: Load and return minimal metadata for the file browser.
        # TODO: Delay loading bulk of the data until file is selected.

        # Limit metadata to scalars and small arrays
        data = self.todict(maxsize=100000)
        # If data['x'] is not a vector or if it was too big, then override
        if self.x.ndim > 1 or len(data['x']) == 0:
            if Intent.isslit(self.intent):
                data['x'] = self.slit1.x.tolist()
            else:
                data['x'] = self.sample.angle_x.tolist()
        return data

    def plot(self, label=None):
        if label is None:
            label = self.name+self.polarization

        from matplotlib import pyplot as plt
        xerr = self.dx if self.angular_resolution is not None else None
        x, dx, xunits, xlabel = self.x, xerr, self.xunits, self.xlabel
        #x, dx, xunits, xlabel = self.detector.angle_x, self.angular_resolution, 'detector angle', 'deg'
        plt.errorbar(x, self.v, yerr=self.dv, xerr=xerr, label=label, fmt='.-')
        plt.xlabel("%s (%s)"%(xlabel, xunits) if xunits else xlabel)
        plt.ylabel("%s (%s)"%(self.vlabel, self.vunits) if self.vunits else self.vlabel)
        if not Intent.isslit(self.intent):
            plt.yscale('log')

    def save(self, filename):
        with open(filename, 'w') as fid:
            fid.write(self.to_column_text()["value"])

    def to_orsopy(self):
        from orsopy import fileio
        info = fileio.Orso.empty()
        info.data_set = f"{self.name}:{self.entry}"
        info.columns = [
            fileio.Column("Qz", "1/angstrom"),
            fileio.Column("R"),
            fileio.ErrorColumn(error_of="R", error_type="uncertainty", value_is="sigma", distribution="gaussian"),
            fileio.ErrorColumn(error_of="Qz", error_type="resolution", value_is="sigma", distribution="gaussian"),
        ]
        data_arrays = [
            self.x,
            self.v,
            self.dv,
            self.dx,
        ]

        instrument_settings = info.data_source.measurement.instrument_settings
        info.data_source.experiment.instrument = str(_s(self.instrument))
        probename = _s(self.probe)
        if probename.endswith('s'):
            # ORSO is expecting "neutron" or "x-ray"
            probename = probename[:-1]
        info.data_source.experiment.probe = probename

        polarization_lookups = {
            "++": "pp",
            "--": "mm",
            "-+": "mp",
            "+-": "pm",
            "+": "po", # assume front polarization but not back
            "-": "mo",
        }
        instrument_settings.polarization = polarization_lookups.get(self.polarization, "unpolarized")

        def pack(ORSO_name, local_name, local_resolution_name, units=None):
            if getattr(self, local_name, None) is not None:
                item = getattr(self, local_name)
                res = getattr(self, local_resolution_name)
                # item and resolution can be either columns or single values,
                # but have to make the same choice for both:
                if np.isscalar(item):
                    item = np.array([item])
                if np.isscalar(res):
                    res = np.array([res])

                item_collapsible = np.allclose(item, item[0]) or len(item) == 1
                res_collapsible = np.allclose(res, res[0]) or len(res) == 1
                collapsed = False
                if item_collapsible and res_collapsible:
                    item = item[0]
                    res = res[0]
                    collapsed = True

                if collapsed:
                    error = fileio.base.ErrorValue(error_value=float(res), error_type="resolution", value_is="sigma", distribution="gaussian")
                    val = fileio.base.Value(magnitude=float(item), unit=units, error=error)
                    setattr(instrument_settings, ORSO_name, val)
                else:
                    info.columns.append(fileio.Column(name=ORSO_name, physical_quantity=ORSO_name, unit=units))
                    info.columns.append(fileio.ErrorColumn(error_of=ORSO_name, error_type="uncertainty", value_is="sigma", distribution="gaussian"))
                    data_arrays.append(np.resize(item, self.points))
                    data_arrays.append(np.resize(res, self.points))
                    val = fileio.base.ValueRange(min=float(min(item)), max=float(max(item)), unit=units)
                    setattr(instrument_settings, ORSO_name, val)

        for ORSO_name, local_name, resolution_name, units in [
            ("wavelength", "Ld", "dL", "angstrom"),
            ("incident_angle", "Ti", "angular_resolution", "degrees"),
        ]:
            pack(ORSO_name, local_name, resolution_name, units)

        ds = fileio.OrsoDataset(info, np.vstack(data_arrays).T)
        return ds

    @exports_ORSO_text("ORSO_text")
    def to_ORSO_text(self):
        orsopy_obj = self.to_orsopy()
        return {
            "name": self.name,
            "entry": self.entry,
            "value": orsopy_obj,
            "file_suffix": ".ort",
        }

    @exports_HDF5(name="ORSO_nexus")
    def to_ORSO_nexus(self):
        from io import BytesIO
        import h5py
        from orsopy.fileio.orso import save_nexus

        orsopy_obj = self.to_orsopy()
        fid = BytesIO()
        save_nexus([orsopy_obj], fid)
        fid.seek(0)
        h5_item = h5py.File(fid, 'r')
        
        return {
            "name": self.name,
            "entry": self.entry,
            "value": h5_item,
            "file_suffix": ".orb",
        }


    # TODO: split refldata in to ReflBase and PointRefl so PSD doesn't inherit column format
    @exports_text("column")
    def to_column_text(self):
        # Note: subclass this for non-traditional reflectometry measurements
        with BytesIO() as fid:  # numpy.savetxt requires a byte stream
            for n in ['name', 'entry', 'polarization']:
                _write_key_value(fid, n, getattr(self, n))
            if self.Ld is not None:
                _write_key_value(fid, "wavelength", self.Ld)
            if self.dL is not None:
                _write_key_value(fid, "wavelength_resolution", self.dL)
            if self.Ti is not None:
                _write_key_value(fid, "angle", self.Ti)
            if self.angular_resolution is not None:
                _write_key_value(fid, "angular_resolution", self.angular_resolution)
            if Intent.isscan(self.intent):
                _write_key_value(fid, "columns", list(self.columns.keys()))
                _write_key_value(fid, "units", [c.get("units", "") for c in self.columns.values()])
                # add column headers
                header_string = "\t".join(list(self.columns.keys())) + "\n"
                fid.write(header_string.encode('utf-8'))
                data_arrays = [
                    self.scan_value[self.scan_label.index(p)] if v.get('is_scan', False)
                    else get_item_from_path(self, p)
                    for p, v in self.columns.items()
                ]
                data_arrays = [np.resize(d, self.points) for d in data_arrays]
                format_string = "\t".join([
                    "%s" if d.dtype.kind in ["S", "U"]
                    else "%.10e"
                    for d in data_arrays
                    ]) + "\n"
                for i in range(self.points):
                    datarow = format_string % tuple([d[i] for d in data_arrays])
                    fid.write(datarow.encode('utf-8'))
                suffix = ".dat"
            else:
                _write_key_value(fid, "columns", [self.xlabel, self.vlabel, "uncertainty", "resolution"])
                _write_key_value(fid, "units", [self.xunits, self.vunits, self.vunits, self.xunits])
                data = np.vstack([self.x, self.v, self.dv, self.dx]).T
                np.savetxt(fid, data, fmt="%.10e")
                suffix = ".refl"
            value = fid.getvalue()

        return {
            "name": self.name,
            "entry": self.entry,
            "file_suffix": suffix,
            "value": value.decode('utf-8'),
        }

    @exports_HDF5(name="NXrefl")
    def to_NXcanSAS(self):
        import h5py
        from io import BytesIO

        fid = BytesIO()
        h5_item = h5py.File(fid, 'w')
        string_dt = h5py.string_dtype(encoding='utf-8')

        #entry_name = metadata.get("entry", "entry")
        nxentry = h5_item.create_group(self.entry)
        nxentry.attrs.update({
            "NX_class": "NXentry",
            "version": "1.0"
        })
        nxentry["definition"] = "NXcanSAS"
        nxentry["run"] = str(self.name)
        nxentry["polarization"] = str(self.polarization)
        nxentry["title"] = ""

        instrument = nxentry.create_group("instrument")
        instrument.attrs.update({
            "canSAS_class": "SASinstrument",
            "NX_class": "NXinstrument"
        })
        instrument["name"] = str(self.instrument)

        refl = nxentry.create_group("columns")
        refl.attrs.update({
            "NX_class": ""
        })
        
        for p, v in self.columns.items():
            if v.get('is_scan', False):
                d = self.scan_value[self.scan_label.index(p)]
            else:
                d = get_item_from_path(self, p)
                #if v.get('')
            if d is not None:
                #d = np.resize(d, self.points)
                refl[p] = d

        sasdetector = instrument.create_group("detector")
        sasdetector.attrs.update({
            "NX_class": "NXdetector"
        })
        sasdetector["name"] = "DETECTOR"
        
        sassample = nxentry.create_group("sassample")
        sassample.attrs.update({
            "NX_class": "NXsample"
        })
        sassample["name"] = str(self.sample.name)
        sassample["description"] = str(self.sample.description)
        
        sassource = instrument.create_group("sassource")
        sassource.attrs.update({
            "canSAS_class": "sassource",
            "NX_class": "NXdetector"
        })
        sassource["radiation"] = "Reactor Neutron Source"

        datagroup = nxentry.create_group("nxrefl")
        datagroup.attrs.update({
            "NX_class": "NXdata",
            "canSAS_class": "REFLdata",
            "signal": "I",
            #"I_axes": "Qz",
            #"Q_indices": 0,
            "timestamp": self.date.isoformat(),
        })

        datagroup["I"] = self.v
        datagroup["I"].attrs["units"] = "arbitrary"
        datagroup["I"].attrs["uncertainties"] = "I_dev"
        datagroup["I_dev"] = np.sqrt(self.dv)
        datagroup["I_dev"].attrs["units"] = "arbitrary"
        datagroup["Q"] = self.Qz
        datagroup["Q"].attrs["units"] = "1/angstrom"
        datagroup["Q"].attrs["uncertainties"] = "Q_dev"
        datagroup["Q_dev"] = self.dQ
        datagroup["Theta"] = self.Td / 2.0
        datagroup["Theta"].attrs["units"] = "degrees"
        datagroup["Theta"].attrs["resolutions"] = "Theta_dev"
        datagroup["Theta"].attrs["resolutions_description"] = "Gaussian"
        datagroup["Theta_dev"] = self.angular_resolution
        datagroup["Theta_dev"].attrs["units"] = "degrees"
        datagroup["Lambda"] = self.detector.wavelength
        datagroup["Lambda"].attrs["resolutions"] = "Lambda_dev"
        datagroup["Lambda"].attrs["units"] = "A"
        datagroup["Lambda_dev"] = self.detector.wavelength_resolution
        datagroup["Lambda_dev"].attrs["units"] = "A"

        # sasaperture = instrument.create_group("sasaperture")
        # sasaperture.attrs.update({
        #     "canSAS_class": "SASaperture",
        #     "NX_class": "NXaperture"
        # })
        # sasaperture["shape"] = "slit"
        # sasaperture["x_gap"] = np.array([0.1], dtype='float')
        # sasaperture["x_gap"].attrs["units"] = "cm"
        # sasaperture["y_gap"] = np.array([5.0], dtype='float')
        # sasaperture["y_gap"].attrs["units"] = "cm"
        
        # wavelength = getattr(self.detector, "wavelength", None)
        # wavelength_resolution = getattr(self.detector, "wavelength_resolution", None)
        # if wavelength is not None:
        #      sassource["incident_wavelength"] = wavelength
        # if wavelength_resolution is not None:
        #     sassource["incident_wavelength_spread"] = wavelength_resolution

        # sassource["incident_wavelength"].attrs["units"] = "A"
        # sassource["incident_wavelength_spread"].attrs["units"] = "A"
        

        sasprocess = nxentry.create_group("sasprocess")
        sasprocess.attrs.update({
            "canSAS_class": "SASprocess",
            "NX_class": "NXprocess"
        })
        sasprocess["name"] = "NIST reductus"

        return {
            "name": self.name,
            "entry": self.entry,
            "file_suffix": ".nxrefl.h5",
            "value": h5_item,
        }

    def get_plottable(self):
        # Note: subclass this for non-traditional reflectometry measurements
        columns = self.columns # {name: {label: str, units: str, errorbars: str}}
        data_arrays = [
            self.scan_value[self.scan_label.index(p)] if v.get('is_scan', False)
            else get_item_from_path(self, p)
            for p, v in columns.items()]
        data_arrays = [np.resize(d, self.points).tolist() for d in data_arrays]
        datas = {c: {"values": d} for c, d in zip(columns.keys(), data_arrays)}
        # add errorbars:
        for k in columns.keys():
            if 'errorbars' in columns[k]:
                #print('errorbars found for column %s' % (k,))
                errorbars = get_item_from_path(self, columns[k]['errorbars'])
                if errorbars is not None:
                    datas[k]["errorbars"] = errorbars.tolist()
                else:
                    print("===> missing errorbars {eb} for {k}".format(eb=columns[k]['errorbars'], k=k))
        name = getattr(self, "name", "default_name")
        entry = getattr(self, "entry", "default_entry")
        series = [{"label": "%s:%s" % (name, entry)}]
        xcol = "x"
        ycol = "v"
        plottable = {
            "type": "nd",
            "title": "%s:%s" % (name, entry),
            "entry": entry,
            "columns": columns,
            "options": {
                "series": series,
                "axes": {
                    "xaxis": {"label": "%s(%s)" % (columns[xcol]["label"], columns[xcol]["units"])},
                    "yaxis": {"label": "%s(%s)" % (columns[ycol]["label"], columns[ycol]["units"])}
                },
                "xcol": xcol,
                "ycol": ycol,
                "errorbar_width": 0
            },
            "datas": datas
        }
        #print(plottable)
        return plottable


class PSDData(ReflData):
    """PSD data for reflectometer"""
    def plot(self, label=None):
        if label is None:
            label = self.name+self.polarization

        from matplotlib import pyplot as plt
        data = np.log(self.v + (self.v == 0))
        plt.pcolormesh(data, label=label)
        plt.xlabel("pixel")
        plt.ylabel("%s (%s)"%(self.xlabel, self.xunits))

    def get_axes(self):
        ny, nx = self.v.shape
        x, xlabel = np.arange(1, nx+1), "pixel"
        if Intent.isslit(self.intent):
            y, ylabel = self.slit1.x, "S1"
        elif Intent.isspec(self.intent):
            y, ylabel = self.Qz_target, "Qz"
        else:
            y, ylabel = np.arange(1, ny+1), "point"
        return (x, xlabel), (y, ylabel)

    def get_plottable(self):
        name = getattr(self, "name", "default_name")
        entry = getattr(self, "entry", "default_entry")
        def limits(v, n):
            low, high = v.min(), v.max()
            delta = (high - low) / max(n-1, 1)
            # TODO: move range cleanup to plotter
            if delta == 0.:
                delta = v[0]/10.
            return low - delta/2, high+delta/2
        data = self.v
        ny, nx = data.shape
        (x, xlabel), (y, ylabel) = self.get_axes()
        #print("data shape", nx, ny)
        xmin, xmax = limits(x, nx)
        ymin, ymax = limits(y, ny)
        # TODO: self.detector.mask
        zmin, zmax = data.min(), data.max()
        # TODO: move range cleanup to plotter
        if zmin <= 0.:
            if (data > 0).any():
                zmin = data[data > 0].min()
            else:
                data[:] = zmin = 1e-10
            if zmin >= zmax:
                zmax = 10*zmin
        dims = {
            "xmin": xmin, "xmax": xmax, "xdim": nx,
            "ymin": ymin, "ymax": ymax, "ydim": ny,
            "zmin": zmin, "zmax": zmax,
        }
        z = data.T.ravel('C').tolist()
        plottable = {
            #'type': '2d_multi',
            #'dims': {'zmin': zmin, 'zmax': zmax},
            #'datasets': [{'dims': dims, 'data': z}],
            'type': '2d',
            'dims': dims,
            'z': [z],
            'entry': entry,
            'title': "%s:%s" % (name, entry),
            'options': {
                'fixedAspect': {
                    'fixAspect': False,
                    'aspectRatio': 1.0,
                },
            },
            'xlabel': xlabel,
            'ylabel': ylabel,
            'zlabel': 'Intensity (I)',
            'ztransform': 'log',
        }
        #print(plottable)
        return plottable

    # TODO: Define export format for partly reduced PSD data.
    @exports_json("json")
    def to_json_text(self):
        name = getattr(self, "name", "default_name")
        entry = getattr(self, "entry", "default_entry")
        return {
            "name": name,
            "entry": entry,
            "file_suffix": ".dat",
            "value": self._toDict(),
            }

    # Kill column writer for now
    def to_column_text(self):
        pass

def get_item_from_path(obj, path):
    """
    Fetch *obj.a.b.c* from path *"a/b/c"*.

    Returns None if path does not exist.
    """
    *head, tail = path.split("/")
    for key in head:
        obj = getattr(obj, key, {})
    return getattr(obj, tail, None)

def _write_key_value(fid, key, value):
    value_str = json.dumps(value, cls=NumpyEncoder)
    if IS_PY3:
        fid.write('# "{0}": {1}\n'.format(key, value_str).encode('utf-8'))
    else:
        fid.write('# "%s": %s\n'%(key, value_str))

def _str(object, indent=4):
    """
    Helper function: document data object by convert attributes listed in
    properties into a string.
    """
    props = [a + "=" + str(getattr(object, a)) for a in object._fields]
    prefix = " "*indent
    return prefix+("\n"+prefix).join(props)

def _toDict(obj, maxsize=np.inf):
    properties = list(getattr(obj, '_fields', ()))
    properties += list(getattr(obj, '_props', ()))
    props = {a: _toDictItem(getattr(obj, a), maxsize=maxsize)
             for a in properties}
    return props

def _toDictItem(obj, maxsize=None):
    if isinstance(obj, np.integer):
        obj = int(obj)
    elif isinstance(obj, np.floating):
        obj = float(obj)
    elif isinstance(obj, np.ndarray):
        obj = obj.tolist() if obj.size < maxsize else [] #[float(obj.min()), float(obj.max())]
    elif isinstance(obj, datetime.datetime):
        obj = [obj.year, obj.month, obj.day, obj.hour, obj.minute, obj.second]
    elif isinstance(obj, (list, tuple)):
        obj = [_toDictItem(a, maxsize) for a in obj]
    return obj

def _fromDict(props):
    # Note: timestamps must have the property named "date"
    for name, value in props.items():
        if isinstance(value, list) and value:
            if all(isinstance(v, (int, float)) for v in value):
                props[name] = np.asarray(value)
        elif name == 'date':
            props[name] = datetime.datetime(*value)
    return props


def _set(object, kw):
    """
    Helper function: distribute the __init__ keyword parameters to
    individual attributes of an object, raising AttributeError if
    the class does not define the given attribute.

    Example:

        def __init__(self, **kw): _set(self,kw)
    """
    for k, v in kw.items():
        # this will fail with an attribute error for incorrect keys
        getattr(object, k)
        setattr(object, k, v)



"""
Based on scalars or numpy vectors, this class allows you to store and
manipulate values+uncertainties, with propagation of gaussian error for
addition, subtraction, multiplication, division, power, exp, log
and trig.  Also includes mean, weighted average, and linear interpolation.

Storage properties are determined by the numbers used to set the value
and uncertainty.  Inputs are coerced to floating point vectors since
numpy does not do automatic type conversion for in-place operations.
If you want single precision or extended precision floats be sure to
set x and variance to one of the those types before calling.

To save memory for huge arrays, the original data is used if it is a
floating point type. In place operations (a*=b, etc.) create at most one
extra copy for each operation. The out-of-place operation c=a*b by contrast
uses four intermediate vectors, so shouldn't be used for huge arrays.
"""
from __future__ import division

__all__ = ['Uncertainty']

from copy import copy

import numpy as np

from . import err1d
from .formatnum import format_uncertainty
from . import wsolve

def float_array(v):
    """
    Convert v to a floating point array if it is not one already.
    """
    if isinstance(v, np.ndarray) and np.issubdtype(v.dtype, np.inexact):
        return v
    if isinstance(v, int):
        return float(v)
    if isinstance(v, float):
        return v
    return np.asarray(v, np.float64)

def _U(x, variance):
    """
    Create an uncertainty object without type-checking.
    """
    #print("creating", x, variance)
    self = Uncertainty.__new__(Uncertainty)
    self._x = x
    self._variance = variance
    return self

# TODO: rename to Measurement and add support for units?
# TODO: C or numba implementation of *, /, **?
# TODO: use __array_function__ for sum, mean, etc.
# TODO: use __array_ufunc__ for all ops rather __mul__, etc.

_UNARY_DISPATCH = {
    np.log: err1d.log,
    np.exp: err1d.exp,
    np.sqrt: err1d.sqrt,
    np.sin: err1d.sin,
    np.cos: err1d.cos,
    np.tan: err1d.tan,
    np.arcsin: err1d.arcsin,
    np.arccos: err1d.arccos,
    np.arctan: err1d.arctan,
    np.negative: lambda a: (-a.x, a.variance),
    np.positive: lambda a: (a.x, a.variance),
    np.absolute: lambda a: (abs(a.x), a.variance),
}
_BINARY_DISPATCH = {
    np.arctan2: err1d.arctan2,
    np.add: err1d.add,
    np.subtract: err1d.sub,
    np.multiply: err1d.mul,
    np.true_divide: err1d.div,
    np.divide: err1d.div,
    np.power: err1d.mul,
    np.multiply: err1d.mul,
    np.multiply: err1d.mul,
}

class Uncertainty(object):
    __slots__ = ('_x', '_variance')
    __array_priority__ = 20.   # force array*uncertainty to use our __rmul__

    # Constructor
    def __init__(self, x, variance=None, dx=None):
        # Assign x first so that its type gets assigned properly.
        self.x = x
        # If standard deviation is given, turn it into variance.
        if dx is not None:
            if variance is not None:
                raise TypeError("Only one of variance and dx can be specified")
            variance = float_array(dx)**2
        # If variance is not given, set it to zero of the correct shape.
        if variance is None:
            variance = np.zeros_like(self.x)
        # Assign variance, possibly coercing its type.
        self.variance = variance

    def __array_ufunc__(self, ufunc, method, *args, **kwargs):
        #print(f"ufunc self:{self} ufunc:{ufunc}, method:{method}, args:{args}, {kwargs}")
        if method != "__call__":
            return NotImplemented
        f = _UNARY_DISPATCH.get(ufunc, None)
        if f is not None:
            a, = args
            return _U(*f(a.x, a.variance))
        #print(f"ufunc {ufunc}")
        f = _BINARY_DISPATCH.get(ufunc, None)
        if f is not None:
            a, b = args
            ax, av = (a.x, a.variance) if isinstance(a, Uncertainty) else (a, 0)
            bx, bv = (b.x, b.variance) if isinstance(b, Uncertainty) else (b, 0)
            return _U(*f(ax, av, bx, bv))
        return NotImplemented 

    def __copy__(self):
        return _U(copy(self.x), copy(self.variance))

    @property
    def x(self):
        return self._x
    @x.setter
    def x(self, v):
        self._x = float_array(v)

    @property
    def variance(self):
        return self._variance
    @variance.setter
    def variance(self, v):
        self._variance = float_array(v)

    @property
    def dtype(self):
        return self.x.dtype
    @property
    def flat(self):
        return iter(self)
    @property
    def ndim(self):
        return self.x.ndim
    @property
    def shape(self):
        return self.x.shape
    @property
    def size(self):
        return self.x.size
    @property
    def T(self):
        return self.transpose()
    # undefined properties:
    #    base, ctypes, data, flags, itemsize, nbytes, strides

    # Comparisons are not well defined for uncertainty objects.  For example,
    # is it true that (3 +/- 0.1) > (2 +/- 10)?  Running a quick simulation,
    # it is only true 54% of the time, so it is neither True nor False.
    # Similarly for equality/inequality.  The following functions are therefore
    # ill-defined:
    #     >, >=, <, <=, ==, !=, bool, all, any, nonzero
    #     argmax, amax, nanmax, maximum, fmax
    #     argmin, amin, nanmin, minimum, fmin
    #     argsort, sort, argpartition, partition, searchsorted
    #     clip, ptp
    # Could perhaps set a p-value for all operations, defaulting to p=0.05,
    # and using two-tailed tests.  That would at least give a partial ordering.
    # Bool, all, and any would involve a comparison against zero.  Maximum
    # likelihood could be used for maximum/minimum, and for the comparisons
    # used for sorting, which amounts to simply using the x value and ignoring
    # the variance.  For now, leave it undefined.

    # getfield, setfield, astype, byteswap, newbyteorder, setflags: use A.x.op

    # TODO: dump, dumps
    # TODO: item, itemset
    # TODO: tobytes, tofile, tolist, tostring

    # choose: index array will not be an uncertainty object, so no choose
    # method, but still may want to implement a choose function.  What happens
    # with np.choose when choices are all uncertainty objects?

    # cumsum mean, sum, std, var: see below
    # TODO: cumprod, prod, dot, trace

    # TODO: round() could truncate x to the number of digits defined by the
    # standard deviation and truncate variance to two digits.

    # TODO: assuming real, imag are identically distributed.  Is that true?
    def conj(self):
        return _U(self.x.conj(), self.variance+0.)
    conjugate = conj
    def real(self):
        return _U(self.x.real, self.variance)
    def imag(self):
        return _U(self.x.imag, self.variance)

    def fill(self, a):
        if isinstance(a, Uncertainty):
            self.x.fill(a.x)
            self.variance.fill(a.variance)
        else:
            self.x.fill(a)
            self.variance.fill(0.)

    # Array shaping/selection operations
    def compress(self, *args, **kw):
        # TODO: if out is supplied, then split it into out.x, out.variance
        return _U(self.x.compress(*args, **kw),
                  self.variance.compress(*args, **kw))
    copy = __copy__
    def diagonal(self, *args, **kw):
        return _U(self.x.diagonal(*args, **kw),
                  self.variance.diagonal(*args, **kw))
    def ravel(self, *args, **kw):
        return _U(self.x.ravel(*args, **kw),
                  self.variance.ravel(*args, **kw))
    def reshape(self, *args, **kw):
        return _U(self.x.reshape(*args, **kw),
                  self.variance.reshape(*args, **kw))
    def repeat(self, *args, **kw):
        return _U(self.x.repeat(*args, **kw),
                  self.variance.repeat(*args, **kw))
    def resize(self, *args, **kw):
        return _U(self.x.resize(*args, **kw),
                  self.variance.resize(*args, **kw))
    def squeeze(self, *args, **kw):
        return _U(self.x.squeeze(*args, **kw),
                  self.variance.squeeze(*args, **kw))
    def swapaxes(self, *args, **kw):
        return _U(self.x.swapaxes(*args, **kw),
                  self.variance.swapaxes(*args, **kw))
    def take(self, *args, **kw):
        return _U(self.x.take(*args, **kw),
                  self.variance.take(*args, **kw))
    def transpose(self, *args, **kw):
        return _U(self.x.transpose(*args, **kw),
                  self.variance.transpose(*args, **kw))
    def view(self, *args, **kw):
        return _U(self.x.view(*args, **kw),
                  self.variance.view(*args, **kw))
    def flatten(self, *args, **kw):
        return _U(self.x.flatten(*args, **kw),
                  self.variance.flatten(*args, **kw))

    # Make standard deviation available
    def _getdx(self):
        return np.sqrt(self.variance)
    def _setdx(self, dx):
        # Direct operation
        #    variance = dx**2
        # Indirect operation to avoid temporary
        self.variance[:] = dx
        self.variance **= 2
    dx = property(_getdx, _setdx, doc="standard deviation")

    def rand(self, size=None):
        return np.random.randn(size=size)*self.dx + self.x

    # Numpy array slicing operations
    def __len__(self):
        return len(self.x)
    def __getitem__(self, key):
        return _U(self.x[key], self.variance[key])
    def __setitem__(self, key, value):
        self.x[key] = value.x
        self.variance[key] = value.variance
    def __delitem__(self, key):
        del self.x[key]
        del self.variance[key]
    def __iter__(self):
        for x, variance in zip(self.x.flat, self.variance.flat):
            yield _U(x, variance)

    # Normal operations: may be of mixed type
    def __add__(self, other):
        if isinstance(other, Uncertainty):
            return _U(*err1d.add(self.x, self.variance, other.x, other.variance))
        else:
            return _U(self.x+other, self.variance+0.)  # Force copy
    def __sub__(self, other):
        if isinstance(other, Uncertainty):
            return _U(*err1d.sub(self.x, self.variance, other.x, other.variance))
        else:
            return _U(self.x-other, self.variance+0.)  # Force copy
    def __mul__(self, other):
        if isinstance(other, Uncertainty):
            return _U(*err1d.mul(self.x, self.variance, other.x, other.variance))
        else:
            return _U(self.x*other, self.variance*other**2)
    def __truediv__(self, other):
        if isinstance(other, Uncertainty):
            return _U(*err1d.div(self.x, self.variance, other.x, other.variance))
        else:
            return _U(self.x/other, self.variance/other**2)
    def __pow__(self, other):
        if isinstance(other, Uncertainty):
            return _U(*err1d.pow2(self.x, self.variance, other.x, other.variance))
        else:
            return _U(*err1d.pow(self.x, self.variance, other))

    # Reverse operations are definitely of mixed type
    def __radd__(self, other):
        return _U(self.x+other, self.variance+0.)  # Force copy
    def __rsub__(self, other):
        return _U(other-self.x, self.variance+0.)  # Force copy
    def __rmul__(self, other):
        return _U(self.x*other, self.variance*other**2)
    def __rtruediv__(self, other):
        return _U(other/self.x, self.variance*other**2/self.x**4)
    def __rpow__(self, other):
        return exp(np.log(other)*self)

    # In-place operations: may not be of mixed type
    # Note that the inplace operations are only inplace for numpy vectors.
    # For scalars, need to assign the new value of the array.
    def __iadd__(self, other):
        if isinstance(other, Uncertainty):
            self._x, self._variance = \
                err1d.add_inplace(self.x, self.variance, other.x, other.variance)
        else:
            self.x += other
        return self
    def __isub__(self, other):
        if isinstance(other, Uncertainty):
            self._x, self._variance = \
                err1d.sub_inplace(self.x, self.variance, other.x, other.variance)
        else:
            self.x -= other
        return self
    def __imul__(self, other):
        if isinstance(other, Uncertainty):
            self._x, self._variance = \
                err1d.mul_inplace(self.x, self.variance, other.x, other.variance)
        else:
            self.x *= other
            self.variance *= other**2
        return self
    def __itruediv__(self, other):
        if isinstance(other, Uncertainty):
            self._x, self._variance = \
                err1d.div_inplace(self.x, self.variance, other.x, other.variance)
        else:
            self.x /= other
            self.variance /= other**2
        return self
    def __ipow__(self, other):
        if isinstance(other, Uncertainty):
            self._x, self._variance = \
                err1d.pow2_inplace(self.x, self.variance, other.x, other.variance)
        else:
            self._x, self._variance = \
                err1d.pow_inplace(self.x, self.variance, other)
        return self

    # always uses true division, even if division is not imported from future
    def __div__(self, other):
        return self.__truediv__(other)
    def __rdiv__(self, other):
        return self.__rtruediv__(other)
    def __idiv__(self, other):
        return self.__itruediv__(other)

    # Unary ops
    def __neg__(self):
        return _U(-self.x, self.variance+0.)  # foce copy
    def __pos__(self):
        return self
    def __abs__(self):
        return _U(np.abs(self.x), self.variance+0.)  # force copy

    def __str__(self):
        #return str(self.x)+" +/- "+str(numpy.sqrt(self.variance))
        if np.isscalar(self.x):
            return format_uncertainty(self.x, self.dx)
        # TODO: don't need to format the entire matrix when printing summary
        formatted = [format_uncertainty(v, dv)
                     for v, dv in zip(self.x.flat, self.dx.flat)]
        with np.printoptions(formatter={'str_kind': lambda s: s}):
            return str(np.array(formatted).reshape(self.x.shape))
    def __repr__(self):
        result = "Uncertainty(%s,%s)"%(repr(self.x), repr(self.variance))
        return result.replace(' ', '').replace('\n', '').replace('array', '')

    #def __format__(self, *args, **kw):
    #    print("calling format with", args, kw)
    #    return str(self)

    # Not implemented
    def __matmul__(self, other): return NotImplemented
    def __rmatmul__(self, other): return NotImplemented
    def __imatmul__(self, other): return NotImplemented

    # Integer operations not defined
    def __floordiv__(self, other): return NotImplemented
    def __mod__(self, other): return NotImplemented
    def __divmod__(self, other): return NotImplemented
    def __lshift__(self, other): return NotImplemented
    def __rshift__(self, other): return NotImplemented
    def __and__(self, other): return NotImplemented
    def __xor__(self, other): return NotImplemented
    def __or__(self, other): return NotImplemented

    def __rfloordiv__(self, other): return NotImplemented
    def __rmod__(self, other): return NotImplemented
    def __rdivmod__(self, other): return NotImplemented
    def __rlshift__(self, other): return NotImplemented
    def __rrshift__(self, other): return NotImplemented
    def __rand__(self, other): return NotImplemented
    def __rxor__(self, other): return NotImplemented
    def __ror__(self, other): return NotImplemented

    def __ifloordiv__(self, other): return NotImplemented
    def __imod__(self, other): return NotImplemented
    def __ilshift__(self, other): return NotImplemented
    def __irshift__(self, other): return NotImplemented
    def __iand__(self, other): return NotImplemented
    def __ixor__(self, other): return NotImplemented
    def __ior__(self, other): return NotImplemented

    def __invert__(self): return NotImplemented  # For ~x
    def __complex__(self): return NotImplemented
    def __int__(self): return NotImplemented
    def __float__(self): return NotImplemented  # don't silently ignore variance
    def __long__(self): return NotImplemented
    def __oct__(self): return NotImplemented
    def __hex__(self): return NotImplemented
    def __index__(self): return NotImplemented
    def __coerce__(self): return NotImplemented

    # Want to support numpy ufunc handling so that np.sin(A)
    # returns Uncertainty(*err1d.sin(A)) when A is an uncertainty object.
    # numpy.sin(x) first calls np.asarray(x) to coerce the input into an
    # array-like shape, then runs through the ufunc on each element of x.
    # This process will not work if __array__ is defined.
    #def __array__(self): return NotImplemented

    def mean(self, axis=None, dtype=None, out=None, keepdims=False, biased=False):
        r"""
        Compute the uncertainty-weighted average of the dataset.

        Use *v.mean(biased=True)* when the variance in the individual points
        is estimated from the data. The result is scaled by the normalized
        $\chi^2$ to correct the bias.[1] Use *v.mean(biased=False)* for the
        variance-weighted average without the $\chi^2$ correction.
        wikipedia page for the weighted arithmetic mean for details.

        [1] https://en.wikipedia.org/wiki/Weighted_arithmetic_mean#Variance_weights
        """
        out_pair = None if out is None else (out.x, out.variance)
        M, varM = err1d.mean(
            self.x, self.variance, biased=biased,
            axis=axis, dtype=dtype, out=out_pair, keepdims=keepdims)
        return _U(M, varM) if out is None else out

    def sum(self, axis=None, dtype=None, out=None, keepdims=False):
        r"""
        Return the sum of an uncertain dataset.

        See numpy.sum for details on the interface.
        """
        out_pair = None if out is None else (out.x, out.variance)
        M, varM = err1d.sum(self.x, self.variance, axis=axis,
                            dtype=dtype, out=out_pair, keepdims=keepdims)
        return _U(M, varM) if out is None else out

    def cumsum(self, axis=None, dtype=None, out=None):
        r"""
        Return the cumulative sum of an uncertain dataset.

        See numpy.cumsum for details on the interface.
        """
        out_pair = None if out is None else (out.x, out.variance)
        M, varM = err1d.cumsum(self.x, self.variance, axis=axis,
                               dtype=dtype, out=out_pair)
        return _U(M, varM) if out is None else out

    def std(self, *args, **kw):
        raise TypeError("Use mean() to compute the mean and variance of uncertainty items")

    def var(self, *args, **kw):
        raise TypeError("Use mean() to compute the mean and variance of uncertainty items")

    def average(self, axis=None, weights=None, returned=False):
        r"""
        Return the weighted average of a data set.

        Note that mean(A) is the same as average(A, weights=1)
        """
        # TODO: explain difference between mean and average
        w, wvar = (
            (weights.x, weights.variance) if isinstance(weights, Uncertainty)
            else (1, 0) if weights is None
            else (weights, 0))
        M, varM = err1d.average(self.x, self.variance, w, wvar, axis=axis)
        return _U(M, varM)

    # TODO: comparisons returning p-value?
    def __eq__(self, other):
        if not isinstance(other, Uncertainty):
            return False
        return (self._x == other._x) & (self._variance == other._variance)

# numpy.method(x) calls x.method() if x has a method attribute, so just use
# the names from the numpy namespace.
sum = np.sum
cumsum = np.cumsum
mean = np.mean
#average = np.average   ## np.average(x) does not call x.average()
log = np.log
exp = np.exp
sqrt = np.sqrt
sin = np.sin
cos = np.cos
tan = np.tan
arcsin = np.arcsin
arccos = np.arccos
arctan = np.arctan
arctan2 = np.arctan2

def average(x, axis=None, weights=None, returned=False):
    r"""
    Return the weighted average of a data set.

    This uses $\sum w_k x_k / \sum w_k$ with gaussian uncertainty propagation
    with $\Delta w_k$ and $\Delta x_k$. Use $w_k = 1 \pm 0$ for a simple
    gaussian average.
    """
    return x.average(axis=axis, weights=weights)


def interp(x, xp, fp, left=None, right=None):
    """
    Linear interpolation of x into points (xk, yk +/- dyk).

    xp is assumed to be in ascending order.

    left is the uncertainty value to return for points before the range of xp,
    or None for the initial value, fp[0].

    right is the uncertainty value to return for points after the range of xp,
    or None for the final value, fp[-1].
    """
    if left is None: left = fp[0]
    if right is None: right = fp[-1]
    left = (left.x, left.variance)
    right = (right.x, right.variance)

    if isinstance(x, np.ndarray) and x.ndim > 0:
        F, varF = err1d.interp(x, xp, fp.x, fp.variance, left, right)
    else:
        F, varF = err1d.interp([x], xp, fp.x, fp.variance, left, right)
        F, varF = F[0], varF[0]
    return _U(F, varF)


def smooth(x, xp, fp, degree=2, span=5):
    """
    Windowed least squares smoothing.
    """
    if span > 2:
        y, dy = wsolve.smooth(x, xp, fp, fp.dx, degree=degree, span=span)
        return _U(y, dy**2)
    else:
        # TODO: smooth with extrapolate, but interp will not.
        return interp(x, xp, fp)

def polyfit(x, y, deg, rcond=None, full=False, w=None, cov=False):
    """
    Like numpy.polyfit, but uses uncertainty as the default weight and returns
    coeffients with uncertainty.

    Does not support *full=True*.
    """
    if full:
        raise NotImplementedError("full=True not supported")

    xvar = getattr(x, 'variance', None)
    if xvar is not None:
        if (xvar != 0.).any():
            raise NotImplementedError("Errors in variables not yet supported")
        xval = x.x
    else:
        xval = x
    # WARNING: numpy docs say use 1/sigma for weight, not 1/sigma^2
    yval, yerr = y.x, y.dx

    if w is None:
        w = 1/yerr

    p, V = np.polyfit(xval, yval, deg, rcond=rcond, full=False, w=w, cov=True)
    pvar = np.diagonal(V, axis1=0, axis2=1)
    coeff = _U(p, pvar)
    return (coeff, V) if cov else coeff

def test():
    a = Uncertainty(5, 3)
    b = Uncertainty(4, 2)
    v = Uncertainty([5, 4], [3, 2])

    # Uscalar op scalar
    z = a+4
    assert z.x == 5+4 and z.variance == 3
    z = a-4
    assert z.x == 5-4 and z.variance == 3
    z = a*4
    assert z.x == 5*4 and z.variance == 3*4**2
    z = a/4
    assert z.x == 5./4 and z.variance == 3./4**2

    # Uvector op scalar
    z = v+4
    assert (z == Uncertainty([5+4, 4+4], [3, 2])).all()
    z = v-4
    assert (z == Uncertainty([5-4, 4-4], [3, 2])).all()
    z = v*4
    assert (z == Uncertainty([5*4, 4*4], [3*4**2, 2*4**2])).all()
    z = v/4
    assert (z == Uncertainty([5/4, 4/4], [3./4**2, 2./4**2])).all()

    # Reverse scalar operations
    z = 4+a
    assert z.x == 4+5 and z.variance == 3
    z = 4-a
    assert z.x == 4-5 and z.variance == 3
    z = 4*a
    assert z.x == 4*5 and z.variance == 3*4**2
    z = 4/a
    assert z.x == 4./5 and abs(z.variance - 3./5**4 * 4**2) < 1e-15

    # Power operations
    z = a**2
    assert z.x == 5**2 and z.variance == 4*3*5**2
    z = a**1
    assert z.x == 5**1 and z.variance == 3
    z = a**0
    assert z.x == 5**0 and z.variance == 0
    z = a**-1
    assert z.x == 5**-1 and abs(z.variance - 3./5**4) < 1e-15

    # Binary operations
    z = a+b
    assert z.x == 5+4 and z.variance == 3+2
    z = a-b
    assert z.x == 5-4 and z.variance == 3+2
    z = a*b
    assert z.x == 5*4 and z.variance == (5**2*2 + 4**2*3)
    z = a/b
    assert z.x == 5./4 and abs(z.variance - (3./5**2 + 2./4**2)*(5./4)**2) < 1e-15

    # ===== Inplace operations =====
    # Scalar operations
    y = a+0; y += 4
    z = a+4
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15
    y = a+0; y -= 4
    z = a-4
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15
    y = a+0; y *= 4
    z = a*4
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15
    y = a+0; y /= 4
    z = a/4
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15

    # Power operations
    y = a+0; y **= 4
    z = a**4
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15

    # Binary operations
    y = a+0; y += b
    z = a+b
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15
    y = a+0; y -= b
    z = a-b
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15
    y = a+0; y *= b
    z = a*b
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15
    y = a+0; y /= b
    z = a/b
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15


    # =============== vector operations ================
    # Slicing
    z = Uncertainty(np.array([1, 2, 3, 4, 5]), np.array([2, 1, 2, 3, 2]))
    assert z[2].x == 3 and z[2].variance == 2
    assert (z[2:4].x == [3, 4]).all()
    assert (z[2:4].variance == [2, 3]).all()
    z[2:4] = Uncertainty(np.array([8, 7]), np.array([4, 5]))
    assert z[2].x == 8 and z[2].variance == 4
    A = Uncertainty(np.array([a.x]*2), np.array([a.variance]*2))
    B = Uncertainty(np.array([b.x]*2), np.array([b.variance]*2))

    # TODO complete tests of copy and inplace operations for vectors and slices.

    # Binary operations
    z = A+B
    assert (z.x == 5+4).all() and (z.variance == 3+2).all()
    z = A-B
    assert (z.x == 5-4).all() and (z.variance == 3+2).all()
    z = A*B
    assert (z.x == 5*4).all() and (z.variance == (5**2*2 + 4**2*3)).all()
    z = A/B
    assert (z.x == 5./4).all()
    assert (abs(z.variance - (3./5**2 + 2./4**2)*(5./4)**2) < 1e-15).all()

    # printing; note that sqrt(3) ~ 1.7
    assert str(Uncertainty(5, 3)) == "5.0(17)"
    assert str(Uncertainty(15, 3)) == "15.0(17)"
    assert str(Uncertainty(151.23356, 0.324185**2)) == "151.23(32)"

    xp = np.array([2., 3.])
    fp = Uncertainty(np.array([3., 1.]), np.array([0.04, 0.16]))
    z = interp(2.4, xp, fp)
    assert abs(z.x-2.2) < 2e-15
    assert abs(z.variance-0.04) < 2e-15
    z = interp([0., 2., 2.4, 3., 4.], xp, fp)
    assert np.linalg.norm(z.x - [3, 3, 2.2, 1, 1]) < 2e-15
    assert np.linalg.norm(z.variance - [0.04, 0.04, 0.04, 0.16, 0.16]) < 2e-15
    xp = np.array([2., 3., 3., 4.])
    fp = Uncertainty(np.array([3., 1., 3., 1.]), np.array([0.04, 0.16, 0.04, 0.16]))
    z = interp([2.5, 3., 3.5], xp, fp)
    assert np.linalg.norm(z.x - [2, 3, 2]) < 2e-15
    assert np.linalg.norm(z.variance - [0.05, 0.04, 0.05]) < 2e-15

    # Make sure polyfit is called with uncertainties
    xp = np.array([1., 2., 3., 4.])
    fp = Uncertainty(np.array([1., 3., 3., 3.]), np.array([0.04, 0.16, 0.04, 0.16]))
    p, C = polyfit(xp, fp, 1, cov=True)
    np_p, np_C = np.polyfit(xp, fp.x, 1, w=1/fp.dx, cov=True)
    assert np.linalg.norm(p.x - np_p) < 2e-15
    assert np.linalg.norm(p.variance - np.diagonal(np_C)) < 2e-15

    # Test numpy ufunc redirection
    assert np.log(a) == _U(*err1d.log(a.x, a.variance))
    assert np.exp(a) == _U(*err1d.exp(a.x, a.variance))
    assert np.sqrt(a) == _U(*err1d.sqrt(a.x, a.variance))
    assert np.sin(a) == _U(*err1d.sin(a.x, a.variance))
    assert np.cos(a) == _U(*err1d.cos(a.x, a.variance))
    assert np.tan(a) == _U(*err1d.tan(a.x, a.variance))
    # arcsin/arccos require values in [-1, 1]
    assert np.arcsin(a/6) == _U(*err1d.arcsin(a.x/6, a.variance/36))
    assert np.arccos(a/6) == _U(*err1d.arccos(a.x/6, a.variance/36))
    assert np.arctan(a) == _U(*err1d.arctan(a.x, a.variance))
    assert np.arctan2(a, b) == _U(*err1d.arctan2(a.x, a.variance, b.x, b.variance))

    # Test array forms
    assert (np.log(v) == _U(*err1d.log(v.x, v.variance))).all()
    assert (np.exp(v) == _U(*err1d.exp(v.x, v.variance))).all()
    assert (np.sqrt(v) == _U(*err1d.sqrt(v.x, v.variance))).all()
    assert (np.sin(v) == _U(*err1d.sin(v.x, v.variance))).all()
    assert (np.cos(v) == _U(*err1d.cos(v.x, v.variance))).all()
    assert (np.tan(v) == _U(*err1d.tan(v.x, v.variance))).all()
    # arcsin/arccos require values in [-1, 1]
    assert (np.arcsin(v/6) == _U(*err1d.arcsin(v.x/6, v.variance/36))).all()
    assert (np.arccos(v/6) == _U(*err1d.arccos(v.x/6, v.variance/36))).all()
    assert (np.arctan(v) == _U(*err1d.arctan(v.x, v.variance))).all()
    assert (np.arctan2(v, v) == _U(*err1d.arctan2(v.x, v.variance, v.x, v.variance))).all()

    # Check vector operations
    s = a+b
    assert np.sum(v) == s
    assert (np.cumsum(v) == _U([a.x, s.x], [a.variance, s.variance])).all()
    # Check in-place cumsum
    c = v.copy()
    np.cumsum(c, out=c)
    assert (np.cumsum(v) == c).all()
    # Mean and average uses err1d
    assert np.mean(v) == _U(*err1d.mean(v.x, v.variance, biased=False))
    assert v.mean(biased=True) == _U(*err1d.mean(v.x, v.variance, biased=True))
    assert v.mean(biased=False) == _U(*err1d.mean(v.x, v.variance, biased=False))
    assert average(v) == _U(*err1d.average(v.x, v.variance, 1, 0))
    assert average(v, weights=2) == _U(*err1d.average(v.x, v.variance, 2, 0))

    # rmul with array - Uarray
    A = Uncertainty([[5, 4], [3, 2]], [[1, 1], [1, 1]])
    B = np.array([3, 4])
    C = B[None, :] * A
    assert C[1, 1] == B[1]*A[1, 1]
    C = A * B[None, :]
    assert C[1, 1] == B[1]*A[1, 1]
    C = A * A
    assert C[1, 1] == A[1, 1]*A[1, 1]

    # Don't care about the format at the moment, but make sure that arrays print
    str(A)

if __name__ == "__main__":
    test()



import os
from pathlib import Path
import shutil

def build_client(
    install_dependencies=False,
    sourcemap=False,
    cleanup=False,
):
    """Build the reductus webview client."""

    # check if npm is installed
    if not shutil.which("npm"):
        raise RuntimeError("npm is not installed. Please install npm.")
    client_folder = (Path(__file__).parent / "webreduce").resolve()
    # check if the node_modules directory exists
    node_modules = client_folder / "node_modules"
    os.chdir(client_folder)
    if install_dependencies or not node_modules.exists():
        print("Installing node modules...")
        os.system("npm install")

    # build the client
    print("Building the webview client...")
    cmd = f"npm run build"
    if sourcemap:
        cmd += " -- --sourcemap"
    os.system(cmd)

    if cleanup:
        print("Cleaning up...")
        shutil.rmtree(node_modules)
        print("node_modules folders removed.")

    print("Done.")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Build the webview client.")
    parser.add_argument("--install-dependencies", action="store_true", help="Install dependencies.")
    parser.add_argument("--sourcemap", action="store_true", help="Generate sourcemaps.")
    parser.add_argument("--cleanup", action="store_true", help="Remove the node_modules directory.")
    args = parser.parse_args()
    build_client(
        install_dependencies=args.install_dependencies,
        sourcemap=args.sourcemap,
        cleanup=args.cleanup,
    )
